#!/usr/bin/env python
#
# VTKH5 Input Output
#
# Last rev: 28/10/2022
from __future__ import print_function, division

import numpy as np, h5py

from ..utils import cr, MPI_RANK, MPI_SIZE, MPI_COMM, mpi_reduce, mpi_bcast

VTKTYPE = np.bytes_('UnstructuredGrid')
VTKVERS = np.array([1,0],np.int32)


def _vtkh5_create_structure(file):
	'''
	Create the basic structure of a VTKH5 file
	'''
	# Create main group
	main = file.create_group('VTKHDF')
	main.attrs['Type']    = VTKTYPE
	main.attrs['Version'] = VTKVERS
	# Create cell data group
	cdata = main.create_group('CellData')
	pdata = main.create_group('PointData')
	fdata = main.create_group('FieldData')
	# Return created groups
	return main, cdata, pdata, fdata

def _vtkh5_connectivity_and_offsets(lnods):
	'''
	Build the offsets array (starting point per each element)

	'''
	# Compute the number of points per cell
	ppcell = np.sum(lnods >= 0,axis=1)
	# Compute the number of zeros per cell
	zpcell = np.sum(lnods < 0,axis=1)
	# Flatten the connectivity array
	lnodsf = lnods.flatten('c')
	# Now build the offsets vector
	offset = np.zeros((ppcell.shape[0]+1,),np.int32)
	offset[1:] = np.cumsum(ppcell) + np.cumsum(zpcell)
	return lnodsf, offset

def _vtkh5_write_mesh_serial(file,xyz,lnods,ltype):
	'''
	Write the mesh and the connectivity to the VTKH5 file.
	'''
	# Create dataset for number of points
	npoints, ndim = xyz.shape
	file.create_dataset('NumberOfPoints',(1,),dtype=int,data=npoints)
	file.create_dataset('Points',(npoints,ndim),dtype=xyz.dtype,data=xyz)
	# Create dataset for number of cells
	lnods, offsets = _vtkh5_connectivity_and_offsets(lnods)
	ncells = ltype.shape[0]
	ncsize = lnods.shape[0]
	file.create_dataset('NumberOfCells',(1,),dtype=int,data=ncells)
	file.create_dataset('NumberOfConnectivityIds',(1,),dtype=int,data=ncsize)
	file.create_dataset('Connectivity',(ncsize,),dtype=int,data=lnods)
	file.create_dataset('Offsets',(ncells+1,),dtype=int,data=offsets)
	file.create_dataset('Types',(ncells,),dtype=np.uint8,data=ltype)
	# Return some parameters
	return npoints, ncells 

def _vtkh5_write_mesh_mpio(file,xyz,lnods,ltype, ptable):
	'''
	Write the mesh and the connectivity to the VTKH5 file.
	'''
	myrank, nparts  = MPI_RANK, MPI_SIZE
	# Create datasets for point data
	npoints      = xyz.shape[0]                               # Number of points of this partition
	npG          = int(mpi_reduce(npoints,op='sum',all=True)) # Total number of points
	npoints_dset = file.create_dataset('NumberOfPoints',(nparts,),dtype=int)
	points_dset  = file.create_dataset('Points',(npG,3),dtype=xyz.dtype)
	# Create datasets for cell data
	ncells, npcells = ltype.shape[0], lnods.shape[1]
	lnods, offsets  = _vtkh5_connectivity_and_offsets(lnods)
	ncsize      = ncells*npcells
	ncG, nsG    = int(mpi_reduce(ncells,op='sum',all=True)), int(mpi_reduce(ncsize,op='sum',all=True))
	ncells_dset = file.create_dataset('NumberOfCells',(nparts,),dtype=int)
	nids_dset   = file.create_dataset('NumberOfConnectivityIds',(nparts,),dtype=int)
	conec_dset  = file.create_dataset('Connectivity',(nsG,),dtype=int)
	offst_dset  = file.create_dataset('Offsets',(ncG+nparts,),dtype=int)
	types_dset  = file.create_dataset('Types',(ncG,),dtype=np.uint8)
	# Each partition writes its own part
	# Point data
	istart, iend = ptable.partition_bounds(myrank,points=True)
	npoints_dset[myrank]       = npoints
	points_dset[istart:iend,:] = xyz
	# Cell data
	istart, iend = ptable.partition_bounds(myrank,points=False)
	ncells_dset[myrank] = ncells
	nids_dset[myrank]   = ncsize
	offst_dset[istart+myrank:iend+(myrank+1)] = offsets
	types_dset[istart:iend] = ltype
	# Connectivity
	istart, iend = ptable.partition_bounds(myrank,ndim=npcells,points=False)
	conec_dset[istart:iend] = lnods
	# Return some parameters
	return npG, ncG

def _vtkh5_link_mesh(file,lname):
	'''
	Create external link mesh to the VTKH5 file.
	'''
	file['NumberOfPoints']          = h5py.ExternalLink(lname,'VTKHDF/NumberOfPoints')
	file['NumberOfCells']           = h5py.ExternalLink(lname,'VTKHDF/NumberOfCells')
	file['NumberOfConnectivityIds'] = h5py.ExternalLink(lname,'VTKHDF/NumberOfConnectivityIds')
	file['Points']                  = h5py.ExternalLink(lname,'VTKHDF/Points')
	file['Connectivity']            = h5py.ExternalLink(lname,'VTKHDF/Connectivity')
	file['Offsets']                 = h5py.ExternalLink(lname,'VTKHDF/Offsets')
	file['Types']                   = h5py.ExternalLink(lname,'VTKHDF/Types')


@cr('vtkh5IO.save_mesh')
def vtkh5_save_mesh(fname,mesh,ptable,mpio=True,mode='w'):
	'''
	Save the mesh component into a VTKH5 file
	'''
	if mpio and not MPI_SIZE == 1:
		vtkh5_save_mesh_mpio(fname,mode,mesh.xyz,mesh.connectivity,mesh.eltype2VTK,ptable)
	else:
		vtkh5_save_mesh_serial(fname,mode,mesh.xyz,mesh.connectivity,mesh.eltype2VTK)

def vtkh5_save_mesh_serial(fname,mode,xyz,lnods,ltype):
	'''
	Save the mesh component into a VTKH5 file (serial)
	'''
	# Open file for writing
	file = h5py.File(fname,mode)
	# Create the file structure
	main, _, _, _ = _vtkh5_create_structure(file)
	# Write the mesh
	_vtkh5_write_mesh_serial(main,xyz,lnods,ltype)
	# Close file
	file.close()

def vtkh5_save_mesh_mpio(fname,mode,xyz,lnods,ltype,ptable):
	'''
	Save the mesh component into a VTKH5 file (serial)
	'''
	# Open file for writing
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	# Create the file structure
	main, _, _, _ = _vtkh5_create_structure(file)
	# Write the mesh
	_vtkh5_write_mesh_mpio(main,xyz,lnods,ltype,ptable)
	# Close file
	file.close()


@cr('vtkh5IO.link_mesh')
def vtkh5_link_mesh(fname,lname,mpio=True,mode='w'):
	'''
	Link the mesh component into a VTKH5 file
	'''
	if mpio and not MPI_SIZE == 1:
		vtkh5_link_mesh_mpio(fname,mode,lname)
	else:
		vtkh5_link_mesh_serial(fname,mode,lname)

def vtkh5_link_mesh_serial(fname,mode,lname):
	'''
	Save the mesh component into a VTKH5 file (serial)
	'''
	# Open file for writing
	file = h5py.File(fname,mode)
	# Create the file structure
	main, _, _, _ = _vtkh5_create_structure(file)
	# Link the mesh
	_vtkh5_link_mesh(main,lname)
	# Close file
	file.close()

def vtkh5_link_mesh_mpio(fname,mode,lname):
	'''
	Save the mesh component into a VTKH5 file (serial)
	'''
	# Open file for writing
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	# Create the file structure
	main, _, _, _ = _vtkh5_create_structure(file)
	# Link the mesh
	_vtkh5_link_mesh(main,lname)
	# Close file
	file.close()


@cr('vtkh5IO.save_field')
def vtkh5_save_field(fname,instant,time,point,varDict,ptable,mpio=True,mode='a'):
	'''
	Save the mesh component into a VTKH5 file
	'''
	if mpio and not MPI_SIZE == 1:
		vtkh5_save_field_mpio(fname,mode,instant,time,point,varDict,ptable)
	else:
		vtkh5_save_field_serial(fname,mode,instant,time,point,varDict)

def vtkh5_save_field_serial(fname,mode,instant,time,point,varDict):
	'''
	Save the field component into a VTKH5 file (serial)
	'''
	# Open file for writing (append to a mesh)
	file = h5py.File(fname,mode)
	main = file['VTKHDF']
	# Write dt and instant as field data
	main['FieldData'].create_dataset('InstantValue',(1,),dtype=int,data=instant)
	main['FieldData'].create_dataset('TimeValue',(1,),dtype=float,data=time)
	# Write the variables
	for var in varDict.keys():
		# Obtain in which group to write
		group = 'PointData' if point else 'CellData'
		# Create and write
		main[group].create_dataset(var,varDict[var].shape,dtype=varDict[var].dtype,data=varDict[var])
	# Close file
	file.close()

def vtkh5_save_field_mpio(fname,mode,instant,time,point,varDict,ptable):
	'''
	Save the mesh component into a VTKH5 file (serial)
	'''
	myrank = MPI_RANK
	# Open file for writing
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	main = file['VTKHDF']
	# Write dt and instant as field data
	main['FieldData'].create_dataset('InstantValue',(1,),dtype=int,data=instant)
	main['FieldData'].create_dataset('TimeValue',(1,),dtype=float,data=time)
	# Create the dictionaries
	dsets = {}
	for var in varDict.keys():
		group   = mpi_bcast('PointData' if point else 'CellData',root=1)
		npG     = int(mpi_reduce(varDict[var].shape[0],op='sum',all=True)) # Total number of points
		nsizeG  = int(mpi_reduce(varDict[var].shape[1] if len(varDict[var].shape) > 1 else 0,op='max',all=True))
		dsets[var] = main[group].create_dataset(var,(npG,) if nsizeG == 0 else (npG,nsizeG) if nsizeG > 0 else (npG,),dtype=varDict[var].dtype)
	# Write the variables
	for var in varDict.keys():
		istart, iend = ptable.partition_bounds(myrank,points=point)
		if len(varDict[var].shape) == 1: # Scalar field
			dsets[var][istart:iend] = varDict[var]
		else: # Vectorial or tensorial field
			dsets[var][istart:iend,:] = varDict[var]
	# Close file
	file.close()


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - truncation.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from libc.math  cimport fabs
from .cfuncs    cimport real, c_svector_norm, c_dvector_norm, c_senergy, c_denergy
from ..utils.cr  import cr


@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef int _scompute_truncation_residual(float[:] S, float r):
	'''
	Compute the truncation residual.
	r must be a float precision (r<1) where:
		- r > 0: target residual
		- r < 0: fraction of cumulative energy to retain
	'''
	cdef int ii, N = 0, nS = S.shape[0]
	cdef float normS, accumulative
	if r > 0:
		normS = c_svector_norm(&S[0],0,nS)
		for ii in range(nS):
			accumulative = c_svector_norm(&S[0],ii,nS)/normS
			if accumulative < r: break
			N += 1
	else:
		r = fabs(r)
		normS = c_svector_norm(&S[0],0,nS)
		accumulative = 0
		for ii in range(nS):
			accumulative += S[ii]/normS
			N += 1		
			if accumulative > r: break
	return N

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef int _dcompute_truncation_residual(double[:] S, double r):
	'''
	Compute the truncation residual.
	r must be a float precision (r<1) where:
		- r > 0: target residual
		- r < 0: fraction of cumulative energy to retain
	'''
	cdef int ii, N = 0, nS = S.shape[0]
	cdef double normS, accumulative
	if r > 0:
		normS = c_dvector_norm(&S[0],0,nS)
		for ii in range(nS):
			accumulative = c_dvector_norm(&S[0],ii,nS)/normS
			if accumulative < r: break
			N += 1
	else:
		r = fabs(r)
		normS = c_dvector_norm(&S[0],0,nS)
		accumulative = 0
		for ii in range(nS):
			accumulative += S[ii]/normS
			N += 1		
			if accumulative > r: break
	return N

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def compute_truncation_residual(real[:] S, real r):
	'''
	Compute the truncation residual.
	r must be a float precision (r<1) where:
		- r > 0: target residual
		- r < 0: fraction of cumulative energy to retain
	'''
	if real is double:
		return _dcompute_truncation_residual(S,r)
	else:
		return _scompute_truncation_residual(S,r)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef float _senergy(float[:,:] A, float[:,:] B):
	'''
	Compute RMSE between X_POD and X
	'''
	cdef int m = A.shape[0], n = B.shape[1]
	cdef float Ek = 0.
	Ek = c_senergy(&A[0,0],&B[0,0],m,n)
	return Ek

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef double _denergy(double[:,:] A, double[:,:] B):
	'''
	Compute RMSE between X_POD and X
	'''
	cdef int m = A.shape[0], n = B.shape[1]
	cdef double Ek = 0.
	Ek = c_denergy(&A[0,0],&B[0,0],m,n)
	return Ek

@cr('math.energy')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def energy(real[:,:] A, real[:,:] B):
	'''
	Compute RMSE between X_POD and X
	'''
	if real is double:
		return _denergy(A,B)
	else:
		return _senergy(A,B)


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# DMD plotting utilities.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from .utils        import extract_modes
from ..utils       import gpu_to_cpu
from ..utils.plots import plotResidual, plotFieldStruct2D, plotSnapshot, plotLayout


def plotMode(Phi, omega, mesh, dset, ivar, pointData=True, modes=np.array([1],np.int32),**kwargs):
	'''
	Plot the real and imaginary parts of a mode
	'''
	Phi, omega = gpu_to_cpu(Phi), gpu_to_cpu(omega)
	# Extract the modes to be plotted
	npoints = mesh.size(pointData)
	Phi_real = extract_modes(Phi,ivar,npoints,real=True,modes=modes)
	Phi_imag = extract_modes(Phi,ivar,npoints,real=False,modes=modes)
	# Add to the dataset
	dset.add_field('PHI_REAL',len(modes),Phi_real)
	dset.add_field('PHI_IMAG',len(modes),Phi_imag)	
	# Loop over the modes
	screenshot = kwargs.pop('screenshot',None)
	off_screen = kwargs.pop('off_screen',False)
	for imode, mode in enumerate(modes):
		if screenshot is not None: kwargs['screenshot'] = screenshot % imode
		plotLayout(mesh,dset,2,1,mode-1,vars=['PHI_REAL','PHI_IMAG'],title='Mode %d St = %.3f' % (mode-1, np.abs(omega[mode-1])/(2*np.pi)),off_screen=off_screen,**kwargs)
	# Remove from dataset
	dset.delete('PHI_REAL')
	dset.delete('PHI_IMAG')


def ritzSpectrum(real, imag, fig = None, ax = None, cmap = None):
	'''
	Given the real and imaginary part of the eigenvalues, plot the Ritz Spectrum together with the unit circle
	'''
	real, imag = gpu_to_cpu(real), gpu_to_cpu(imag)
	if fig is None:
		fig = plt.figure(figsize=(8,6),dpi=100)
	if ax is None:
		ax = fig.add_subplot(1,1,1)
	if cmap is None:
		cmap = 'r'
	theta = np.array(range(101))*2*np.pi/100
	ax.plot(np.cos(theta), np.sin(theta), c = 'k')
	ax.scatter(real, imag, c = cmap)
	ax.axis('equal')
	ax.set(xlabel = r'$\mu_{Re}$', ylabel = r'$\mu_{Imag}$', title = 'Ritz spectrum')
	return fig, ax

def amplitudeFrequency(omega, amplitude, fig = None, ax = None, cmap = None, mark = None, norm = False):
	'''
	Given the frequency and the amplitude of the DMD modes, plot the amplitude against the Strouhal number
	'''
	omega, amplitude = gpu_to_cpu(omega), gpu_to_cpu(amplitude)
	if fig is None:
		fig = plt.figure(figsize=(8,6),dpi=100)
	if ax is None:
		ax = fig.add_subplot(1,1,1)
	if cmap is None:
		cmap = 'r'
	if mark is None:
		mark = 'X'
	if norm is True:
		amplitude = np.abs(amplitude)/np.max(np.abs(amplitude))
	else:
		amplitude = np.abs(amplitude)
	ax.scatter(omega/(2*np.pi), amplitude, c = cmap, marker = mark)
	ax.set(xlabel = 'St', ylabel = 'Amplitude', title = 'Amplitude vs Frequency of the DMD Modes')
	ax.set_yscale('log')
	return fig, ax

def dampingFrequency(omega, delta, fig = None, ax = None, cmap = None, mark = None):
	'''
	Given the frequency and the damping ratio of the DMD modes, plot the amplitude against the Strouhal number
	'''
	omega, delta = gpu_to_cpu(omega), gpu_to_cpu(delta)
	if fig is None:
		fig = plt.figure(figsize=(8,6),dpi=100)
	if ax is None:
		ax = fig.add_subplot(1,1,1)
	if cmap is None:
		cmap = 'r'
	if mark is None:
		mark = 'X'
	ax.scatter(omega/(2*np.pi), np.abs(delta), c = cmap, marker = mark)
	ax.set(xlabel = 'St', ylabel = 'Damping Ratio', title = 'Damping ratio vs Frequency of the DMD Modes')
	ax.set_yscale('log')
	return fig, ax



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for SPOD.
#
# Last rev: 02/09/2022
from __future__ import print_function

cimport cython
cimport numpy as np

import numpy as np

#from libc.complex  cimport creal, cimag
cdef extern from "<complex.h>" nogil:
	float  complex I
	# Decomposing complex values
	float cimagf(float complex z)
	float crealf(float complex z)
	double cimag(double complex z)
	double creal(double complex z)
cdef double complex J = 1j
from libc.stdlib     cimport malloc, free
from libc.string     cimport memcpy, memset
from libc.math       cimport pow, floor, ceil, log2, sqrt
from ..vmmath.cfuncs cimport real, real_complex
from ..vmmath.cfuncs cimport c_ssort, c_svector_mean, c_stemporal_mean, c_ssubtract_mean, c_shammwin, c_sfft1D, c_ctsqr_svd
from ..vmmath.cfuncs cimport c_dsort, c_dvector_mean, c_dtemporal_mean, c_dsubtract_mean, c_dhammwin, c_dfft1D, c_ztsqr_svd

from ..utils.cr       import cr, cr_start, cr_stop


@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _sfft(np.complex64_t *out, float *Xf, float winWeight, int nDFT, int nf):
	cdef int i
	cdef float fact = winWeight/nDFT
	cdef np.complex64_t *Yf
	# Allocate memory
	Yf = <np.complex64_t*>malloc(nDFT*sizeof(np.complex64_t))
	# Compute FFT
	c_sfft1D(Yf,Xf,nDFT)
	# Multiply by a factor
	for i in range(nf):
		out[i] = fact*Yf[i]
	# Free memory
	free(Yf)

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _dfft(np.complex128_t *out, double *Xf, double winWeight, int nDFT, int nf):
	cdef int i
	cdef double fact = winWeight/nDFT
	cdef np.complex128_t *Yf
	# Allocate memory
	Yf = <np.complex128_t*>malloc(nDFT*sizeof(np.complex128_t))
	# Compute FFT
	c_dfft1D(Yf,Xf,nDFT)
	# Multiply by a factor
	for i in range(nf):
		out[i] = fact*Yf[i]
	# Free memory
	free(Yf)

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _ssort(float *L, float *P, float *f, int M, int nBlks, int nf):
	cdef int ii, jj, index
	cdef float tmp
	cdef float *v
	cdef int *idx
	v   = <float*>malloc(nf*sizeof(float))
	idx = <int*>malloc(nf*sizeof(int))
	# Copy to v
	for ii in range(nf):
		v[ii] = L[nBlks*ii + 0]
	# Obtain index array that sorts v
	c_ssort(v,idx,nf)
	free(v)
	# Reorder L, P and f arrays
	for ii in range(nf):
		index = nf-idx[ii]-1 # Reversed order
		# Swap f
		tmp      = f[index]
		f[index] = f[ii]
		f[ii]    = tmp
		# Swap L
		for jj in range(nBlks):
			tmp                 = L[nBlks*index + jj]
			L[nBlks*index + jj] = L[nBlks*ii + jj]
			L[nBlks*ii + jj]    = tmp
		# Swap P
		for jj in range(M*nBlks):
			tmp              = P[nf*jj + index]
			P[nf*jj + index] = P[nf*jj + ii]
			P[nf*jj + ii]    = tmp			
	free(idx)

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
cdef void _dsort(double *L, double *P, double *f, int M, int nBlks, int nf):
	cdef int ii, jj, index
	cdef double tmp
	cdef double *v
	cdef int *idx
	v   = <double*>malloc(nf*sizeof(double))
	idx = <int*>malloc(nf*sizeof(int))
	# Copy to v
	for ii in range(nf):
		v[ii] = L[nBlks*ii + 0]
	# Obtain index array that sorts v
	c_dsort(v,idx,nf)
	free(v)
	# Reorder L, P and f arrays
	for ii in range(nf):
		index = nf-idx[ii]-1 # Reversed order
		# Swap f
		tmp      = f[index]
		f[index] = f[ii]
		f[ii]    = tmp
		# Swap L
		for jj in range(nBlks):
			tmp                 = L[nBlks*index + jj]
			L[nBlks*index + jj] = L[nBlks*ii + jj]
			L[nBlks*ii + jj]    = tmp
		# Swap P
		for jj in range(M*nBlks):
			tmp              = P[nf*jj + index]
			P[nf*jj + index] = P[nf*jj + ii]
			P[nf*jj + ii]    = tmp			
	free(idx)


## SPOD run method
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _srun(float[:,:] X, float[:] t, int nDFT, int nolap, int remove_mean):
	'''
	Run SPOD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,nt]: data matrix
		- dt:                timestep between adjacent snapshots
		- npwin:             number of points in each window (0 will set default value: ~10% nt)
		- nolap:             number of overlap points between windows (0 will set default value: 50% nwin)
		- remove_mean:       whether or not to remove the mean flow

	Returns:
		- L:  modal energy spectra.
		- P:  SPOD modes, whose spatial dimensions are identical to those of X.
		- f:  frequency vector.
	''' 
	cdef int i, iblk, ifreq, ip, i0, nBlks, nf, M = X.shape[0], N = X.shape[1]
	cdef float winWeight, dt = t[1] - t[0]

	cdef float *window
	cdef float *X_mean
	cdef float *Xf
	cdef float *Y
	cdef float *S
	cdef np.complex64_t *qk
	cdef np.complex64_t *Q
	cdef np.complex64_t *U
	cdef np.complex64_t *V

	# Output arrays
	cdef np.ndarray[np.float32_t,ndim=2] L
	cdef np.ndarray[np.float32_t,ndim=2] P
	cdef np.ndarray[np.float32_t,ndim=1] f

	# Deal with the window gain
	if nDFT == 0: 
		nDFT = <int>(pow(2,floor(log2(N/10))))
	# Allocate vector
	window = <float*>malloc(nDFT*sizeof(float))
	# Compute Hamming window
	c_shammwin(window,nDFT)

	# Correction for FFT window gain
	winWeight = 1.0/c_svector_mean(window,0,nDFT)

	if nolap == 0:
		nolap = <int>(floor(nDFT/2))

	nBlks = <int>(floor((N-nolap)/(nDFT-nolap)))

	# Remove temporal mean
	Y = <float*>malloc(M*N*sizeof(float))
	if remove_mean:
		cr_start('SPOD.temporal_mean',0)
		X_mean = <float*>malloc(M*sizeof(float))
		# Compute temporal mean
		c_stemporal_mean(X_mean,&X[0,0],M,N)
		# Compute substract temporal mean
		c_ssubtract_mean(Y,&X[0,0],X_mean,M,N)
		free(X_mean)
		cr_stop('SPOD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],M*N*sizeof(float))

	# Set frequency axis
	nf = <int>(ceil(nDFT/2)) + 1

	f = np.zeros((nf,)       ,dtype=np.float)
	L = np.zeros((nf,nBlks)  ,dtype=np.float)
	P = np.zeros((M*nBlks,nf),dtype=np.float)

	# Set frequency axis
	for i in range(nf):
		f[i] = <float>(i)/dt/<float>(nDFT)

	# Allocate memory
	Xf = <float*>malloc(nDFT*sizeof(float))
	qk = <np.complex64_t*>malloc(M*nf*sizeof(np.complex64_t))
	Q  = <np.complex64_t*>malloc(M*nf*nBlks*sizeof(np.complex64_t))
	cr_start('SPOD.fft',0)
	for iblk in range(nBlks):
		i0 = iblk*(nDFT - nolap)
		for ip in range(M):
			# Populate Xf
			for i in range(nDFT):
				Xf[i] = Y[N*ip + (i + i0)] * window[i]
			# FFT on Xf
			_sfft(&qk[nf*ip],Xf,winWeight,nDFT,nf)
			# Multiply qk
			for i in range(1,nf-1):
				qk[nf*ip + i] *= <float>(2.0)
		# Populate Q
		for ip in range(M):
			for i in range(nf):
				Q[nf*nBlks*ip + nBlks*i + iblk] = qk[nf*ip + i]
	cr_stop('SPOD.fft',0)

	free(qk)
	free(window)
	free(Xf)

	# Allocate memory
	qf = <np.complex64_t*>malloc(M*nBlks*sizeof(np.complex64_t))
	U  = <np.complex64_t*>malloc(M*nBlks*sizeof(np.complex64_t))
	S  = <float*>malloc(M*nBlks*sizeof(float))
	V  = <np.complex64_t*>malloc(M*nBlks*sizeof(np.complex64_t))

	cr_start('SPOD.SVD',0)
	for ifreq in range(nf):
		# Load block in qf
		for i in range(M):
			for iblk in range(nBlks):
				winWeight = sqrt(nBlks) # reused variable
				qf[nBlks*i + iblk] = crealf(Q[nf*nBlks*i + nBlks*ifreq + iblk])/winWeight + cimagf(Q[nf*nBlks*i + nBlks*ifreq + iblk])/winWeight
		# Run SVD
		c_ctsqr_svd(U,S,V,qf,M,nBlks)
		# Store P
		for i in range(M):
			for iblk in range(nBlks):
				P[i + M*iblk,ifreq] = crealf(U[nBlks*i + iblk])
		# Store L
		for iblk in range(nBlks):
			L[ifreq,iblk] = S[iblk]*S[iblk]
	cr_stop('SPOD.SVD',0)

	free(qf)
	free(Q)
	free(U)
	free(S)
	free(V)

	# Sort
	cr_start('SPOD.sort',0)
	_ssort(&L[0,0],&P[0,0],&f[0],M,nBlks,nf)
	cr_stop('SPOD.sort',0)

	return L, P, f

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _drun(double[:,:] X, double[:] t, int nDFT, int nolap, int remove_mean):
	'''
	Run SPOD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,nt]: data matrix
		- dt:                timestep between adjacent snapshots
		- npwin:             number of points in each window (0 will set default value: ~10% nt)
		- nolap:             number of overlap points between windows (0 will set default value: 50% nwin)
		- remove_mean:       whether or not to remove the mean flow

	Returns:
		- L:  modal energy spectra.
		- P:  SPOD modes, whose spatial dimensions are identical to those of X.
		- f:  frequency vector.
	''' 
	cdef int i, iblk, ifreq, ip, i0, nBlks, nf, M = X.shape[0], N = X.shape[1]
	cdef double winWeight, dt = t[1] - t[0]

	cdef double *window
	cdef double *X_mean
	cdef double *Xf
	cdef double *Y
	cdef np.complex128_t *qk
	cdef np.complex128_t *Q
	cdef np.complex128_t *U
	cdef double *S
	cdef np.complex128_t *V

	# Output arrays
	cdef np.ndarray[np.double_t,ndim=2] L
	cdef np.ndarray[np.double_t,ndim=2] P
	cdef np.ndarray[np.double_t,ndim=1] f

	# Deal with the window gain
	if nDFT == 0: 
		nDFT = <int>(pow(2,floor(log2(N/10))))
	# Allocate vector
	window = <double*>malloc(nDFT*sizeof(double))
	# Compute Hamming window
	c_dhammwin(window,nDFT)

	# Correction for FFT window gain
	winWeight = 1.0/c_dvector_mean(window,0,nDFT)

	if nolap == 0:
		nolap = <int>(floor(nDFT/2))

	nBlks = <int>(floor((N-nolap)/(nDFT-nolap)))

	# Remove temporal mean
	Y = <double*>malloc(M*N*sizeof(double))
	if remove_mean:
		cr_start('SPOD.temporal_mean',0)
		X_mean = <double*>malloc(M*sizeof(double))
		# Compute temporal mean
		c_dtemporal_mean(X_mean,&X[0,0],M,N)
		# Compute substract temporal mean
		c_dsubtract_mean(Y,&X[0,0],X_mean,M,N)
		free(X_mean)
		cr_stop('SPOD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],M*N*sizeof(double))

	# Set frequency axis
	nf = <int>(ceil(nDFT/2)) + 1

	f = np.zeros((nf,)       ,dtype=np.double)
	L = np.zeros((nf,nBlks)  ,dtype=np.double)
	P = np.zeros((M*nBlks,nf),dtype=np.double)

	# Set frequency axis
	for i in range(nf):
		f[i] = <double>(i)/dt/<double>(nDFT)

	# Allocate memory
	Xf = <double*>malloc(nDFT*sizeof(double))
	qk = <np.complex128_t*>malloc(M*nf*sizeof(np.complex128_t))
	Q  = <np.complex128_t*>malloc(M*nf*nBlks*sizeof(np.complex128_t))
	cr_start('SPOD.fft',0)
	for iblk in range(nBlks):
		i0 = iblk*(nDFT - nolap)
		for ip in range(M):
			# Populate Xf
			for i in range(nDFT):
				Xf[i] = Y[N*ip + (i + i0)] * window[i]
			# FFT on Xf
			_dfft(&qk[nf*ip],Xf,winWeight,nDFT,nf)
			# Multiply qk
			for i in range(1,nf-1):
				qk[nf*ip + i] *= 2.0
		# Populate Q
		for ip in range(M):
			for i in range(nf):
				Q[nf*nBlks*ip + nBlks*i + iblk] = qk[nf*ip + i]
	cr_stop('SPOD.fft',0)

	free(qk)
	free(window)
	free(Xf)

	# Allocate memory
	qf = <np.complex128_t*>malloc(M*nBlks*sizeof(np.complex128_t))
	U  = <np.complex128_t*>malloc(M*nBlks*sizeof(np.complex128_t))
	S  = <double*>malloc(M*nBlks*sizeof(double))
	V  = <np.complex128_t*>malloc(M*nBlks*sizeof(np.complex128_t))

	cr_start('SPOD.SVD',0)
	for ifreq in range(nf):
		# Load block in qf
		for i in range(M):
			for iblk in range(nBlks):
				qf[nBlks*i + iblk] = Q[nf*nBlks*i + nBlks*ifreq + iblk]/sqrt(nBlks)
		# Run SVD
		c_ztsqr_svd(U,S,V,qf,M,nBlks)
		# Store P
		for i in range(M):
			for iblk in range(nBlks):
				P[i + M*iblk,ifreq] = creal(U[nBlks*i + iblk])
		# Store L
		for iblk in range(nBlks):
			L[ifreq,iblk] = S[iblk]*S[iblk]
	cr_stop('SPOD.SVD',0)

	free(qf)
	free(Q)
	free(U)
	free(S)
	free(V)

	# Sort
	cr_start('SPOD.sort',0)
	_dsort(&L[0,0],&P[0,0],&f[0],M,nBlks,nf)
	cr_stop('SPOD.sort',0)

	return L, P, f

@cr('SPOD.run')
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def run(real[:,:] X, real[:] t, int nDFT=0, int nolap=0, int remove_mean=True):
	'''
	Run SPOD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,nt]: data matrix
		- dt:                timestep between adjacent snapshots
		- npwin:             number of points in each window (0 will set default value: ~10% nt)
		- nolap:             number of overlap points between windows (0 will set default value: 50% nwin)
		- remove_mean:       whether or not to remove the mean flow

	Returns:
		- L:  modal energy spectra.
		- P:  SPOD modes, whose spatial dimensions are identical to those of X.
		- f:  frequency vector.
	''' 
	if real is double:
		return _drun(X,t,nDFT,nolap,remove_mean)
	else:
		return _srun(X,t,nDFT,nolap,remove_mean)


#!/usr/bin/env python
#
# PYLOM Testsuite
# Build cylinder dataset for the testsuite
#
# Last revision: 03/08/2021
from __future__ import print_function, division

import numpy as np
from scipy.io import loadmat

import pyLOM


## Parameters
MATFILE = '../DATA/CYLINDER_ALL.mat'
OUTFILE = './DATA/CYLINDER.h5'
DT      = 0.2
DIMSX   = -1., 8.
DIMSY   = -2., 2.


## Load MAT file
mat = loadmat(MATFILE)


## Create the mesh
nx, ny = int(mat['ny'])+1,int(mat['nx'])+1 # wrong in the mat file
mesh = pyLOM.Mesh.new_struct2D(nx,ny,None,None,DIMSX,DIMSY) 
print(mesh)


## Create partition table
ptable = pyLOM.PartitionTable.new(1,mesh.ncells,mesh.npoints)
mesh.partition_table = ptable
print(ptable)

# Build time instants
time = DT*np.arange(mat['UALL'].shape[1])

# Build velocity as a 2D array
VELOC = np.zeros((2*mesh.ncells,time.shape[0]),dtype=np.double)
VELOC[:2*mesh.ncells:2,:]  = np.ascontiguousarray(mat['UALL'].astype(np.double))
VELOC[1:2*mesh.ncells:2,:] = np.ascontiguousarray(mat['VALL'].astype(np.double))

VORTI = np.zeros((1*mesh.ncells,time.shape[0]),dtype=np.double)
VORTI[:,:] = np.ascontiguousarray(mat['VORTALL'].astype(np.double))

VELOX = np.zeros((1*mesh.ncells,time.shape[0]),dtype=np.double)
VELOX[:,:] = np.ascontiguousarray(mat['UALL'].astype(np.double))


## Create dataset for pyLOM
# Select a limited number of time instants
nsample = 10
d = pyLOM.Dataset(xyz=mesh.xyzc, ptable=ptable, order=mesh.cellOrder, point=False,
	# Add the time as the only variable
	vars  = {'time':{'idim':0,'value':time[::nsample].copy()}},
	# Now add all the arrays to be stored in the dataset
	# It is important to convert them as C contiguous arrays
	VELOC = {'ndim':2,'value':VELOC[:,::nsample].copy()},
	VELOX = {'ndim':1,'value':VELOX[:,::nsample].copy()},
	VORTI = {'ndim':1,'value':VORTI[:,::nsample].copy()},
)
print(d)

mesh.save(OUTFILE) # Store the mesh
d.save(OUTFILE)    # Store dataset


pyLOM.cr_info()


#!/usr/bin/env python
#
# pyLOM, IO
#
# H5 Input Output
#
# Last rev: 31/07/2021
from __future__ import print_function, division

import os, numpy as np, h5py

from ..partition_table import PartitionTable
from ..mesh            import MTYPE2ID, ID2MTYPE
from ..utils           import cr, MPI_COMM, MPI_RANK, MPI_SIZE, worksplit, writesplit, is_rank_or_serial, mpi_reduce, mpi_gather, raiseError


PYLOM_H5_VERSION = (3,0)


def h5_save_partition(file,ptable):
	'''
	Save a partition table inside an HDF5 file
	'''
	# Create a group for the mesh
	group = file.create_group('PARTITIONS')
	group.create_dataset('NSubD',(1,),dtype='i4',data=ptable.n_partitions)
	group.create_dataset('Ids',(ptable.n_partitions,),dtype='i4',data=ptable.Ids)
	group.create_dataset('Elements',(ptable.n_partitions,),dtype='i4',data=ptable.Elements)
	group.create_dataset('Points',(ptable.n_partitions,),dtype='i4',data=ptable.Points)

def h5_load_partition(file):
	'''
	Load a partition table inside an HDF5 file
	'''
	# Load file
	if not 'PARTITIONS' in file.keys(): raiseError('No partition table stored in dataset!')
	nparts   = int(file['PARTITIONS']['NSubD'][0])
	ids      = np.array(file['PARTITIONS']['Ids'][:])
	elements = np.array(file['PARTITIONS']['Elements'][:])
	points   = np.array(file['PARTITIONS']['Points'][:])
	# Return partition class
	return PartitionTable(nparts,ids,elements,points)

def h5_save_meshes(file,mtype,xyz,conec,eltype,cellO,pointO,ptable):
	'''
	Save the mesh inside the HDF5 file
	'''
	# Save the mesh type
	file.create_dataset('type',(1,),dtype='i4',data=MTYPE2ID[mtype])
	# Write the total number of cells and the total number of points
	# Assume we might be dealing with a parallel mesh
	ndim     = xyz.shape[1]
	nnodcell = conec.shape[1]
	npointG  = mpi_reduce(xyz.shape[0],op='sum',all=True)
	ncellG   = mpi_reduce(eltype.shape[0],op='sum',all=True)
	if ptable.has_master: 
		npointG -= 1
		ncellG  -= 1
	file.create_dataset('npoints',(1,),dtype='i4',data=npointG)
	file.create_dataset('ncells' ,(1,),dtype='i4',data=ncellG)
	# Create the rest of the datasets for parallel storage
	dxyz   = file.create_dataset('xyz',(npointG,ndim),dtype=xyz.dtype)
	dconec = file.create_dataset('connectivity',(ncellG,nnodcell),dtype='i4')
	deltyp = file.create_dataset('eltype',(ncellG,),dtype='u1')
	dcellO = file.create_dataset('cellOrder',(ncellG,),dtype='i4')
	dpoinO = file.create_dataset('pointOrder',(npointG,),dtype='i4')
	# Skip master if needed
	if ptable.has_master and MPI_RANK == 0: return None, None, None
	# Point dataset
	# Compute start and end of read, node data
	istartp, iend = ptable.partition_bounds(MPI_RANK,points=True)
	dxyz[istartp:iend,:]  = xyz
	dpoinO[istartp:iend]  = pointO
	# Compute start and end of read, cell data
	istart, iend = ptable.partition_bounds(MPI_RANK,points=False)
	dconec[istart:iend,:] = conec + istartp
	deltyp[istart:iend]   = eltype
	dcellO[istart:iend]   = cellO

def h5_save_meshes_nopartition(file,mtype,xyz,conec,eltype,cellO,pointO,ptable):
	'''
	Save the mesh inside the HDF5 file
	'''
	# Save the mesh type
	file.create_dataset('type',(1,),dtype='i4',data=MTYPE2ID[mtype])
	# Write the total number of cells and the total number of points
	# Assume we might be dealing with a parallel mesh
	ndim     = xyz.shape[1]
	nnodcell = conec.shape[1]
	npointG  = mpi_reduce(pointO.max() if pointO.shape[0] > 0 else 0,op='max',all=True) + 1
	ncellG   = mpi_reduce(cellO.max() if cellO.shape[0] > 0 else 0,op='max',all=True)  + 1
	file.create_dataset('npoints',(1,),dtype='i4',data=npointG)
	file.create_dataset('ncells' ,(1,),dtype='i4',data=ncellG)
	# Create the rest of the datasets for parallel storage
	dxyz   = file.create_dataset('xyz',(npointG,ndim),dtype=xyz.dtype)
	dconec = file.create_dataset('connectivity',(ncellG,nnodcell),dtype='i4')
	deltyp = file.create_dataset('eltype',(ncellG,),dtype='u1')
	dcellO = file.create_dataset('cellOrder',(ncellG,),dtype='i4')
	dpoinO = file.create_dataset('pointOrder',(npointG,),dtype='i4')
	# Skip master if needed
	if ptable.has_master and MPI_RANK == 0: return None, None, None
	# Point dataset
	# Get the position where the points should be stored
	inods,idx = np.unique(pointO,return_index=True)
	dxyz[inods,:] = xyz[idx,:]
	dpoinO[inods] = pointO[idx]
	# Compute start and end of read, cell data
	istart, iend = ptable.partition_bounds(MPI_RANK,points=False)
	dconec[istart:iend,:] = pointO[conec] if pointO.shape[0] > 0 else conec
	deltyp[istart:iend]   = eltype
	dcellO[istart:iend]   = cellO

def h5_load_meshes_size(file):
	'''
	Load only the number of cells and points for the partition
	'''
	# If the mesh is present read the size
	npoints = int(file['npoints'][0])
	ncells  = int(file['ncells'][0])
	return npoints, ncells

def h5_load_meshes(file,ptable,repart):
	'''
	Load the mesh inside the HDF5 file
	'''
	# Read mesh type
	mtype  = ID2MTYPE[int(file['type'][0])]
	# Read cell related variables
	istart, iend = ptable.partition_bounds(MPI_RANK,points=False)
	conec  = np.array(file['connectivity'][istart:iend,:],np.int32)
	eltype = np.array(file['eltype'][istart:iend],np.int32) 
	cellO  = np.array(file['cellOrder'][istart:iend],np.int32)
	cellO  = np.arange(istart, iend, 1) # Fix IO for clipped datasets in pyQvarsi
	# Read point related variables
	if repart:
		# Warning! Repartition will only work if the input file is serial
		# i.e., it does not have any repeated nodes, otherwise it wont work
		ptable.create_partition_points(conec)
		inods  = ptable.partition_points(1)
	else:
		istart, iend = ptable.partition_bounds(MPI_RANK,points=True)
		inods = np.arange(istart,iend,dtype=np.int32)
	xyz    = np.array(file['xyz'][inods,:],file['xyz'].dtype) 
	pointO = np.array(file['pointOrder'][inods],np.int32)
	# Fix the connectivity to start at zero
	conec2 = -np.ones_like(conec).flatten()# This is a 1D array of -1 of the size of our connectivity
	conec2[conec.flatten() >= 0] = np.searchsorted(pointO, conec[conec >= 0].flatten()) # Search only the positive values
	conec = conec2.reshape(conec.shape).astype(np.int32) # Reshape the connectivity to its original format
	print(conec,conec[conec<0])
	# Return
	return mtype, xyz, conec, eltype, cellO, pointO

def h5_save_points(file,xyz,order,ptable,point):
	'''
	Save the points inside the HDF5 file
	'''
	npointG = mpi_reduce(xyz.shape[0] if not np.any(np.isnan(xyz)) else 0,op='sum',all=True)
	ndim    = xyz.shape[1]
	if ptable.has_master: npointG -= 1
	file.create_dataset('pointData',(1,),dtype='i4',data=point)
	file.create_dataset('npoints',(1,),dtype='i4',data=npointG)
	# Create the rest of the datasets for parallel storage
	dxyz   = file.create_dataset('xyz',(npointG,ndim),dtype=xyz.dtype)
	dpoinO = file.create_dataset('order',(npointG,),dtype='i4')
	# Skip master if needed
	if ptable.has_master and MPI_RANK == 0: return None, None, None
	# Compute start and end of read, node data
	istart, iend = ptable.partition_bounds(MPI_RANK,points=point)
	dxyz[istart:iend,:] = xyz
	dpoinO[istart:iend] = order
	return None, None, None

def h5_save_points_nopartition(file,xyz,order,ptable,point):
	'''
	Save the points inside the HDF5 file
	'''
	# Assume we might be dealing with a parallel mesh
	npointG = mpi_reduce(order.max() if order.shape[0] > 0 else 0,op='max',all=True) + 1
	ndim    = xyz.shape[1]
	file.create_dataset('pointData',(1,),dtype='i4',data=point)
	file.create_dataset('npoints',(1,),dtype='i4',data=npointG)
	# Create the rest of the datasets for parallel storage
	dxyz   = file.create_dataset('xyz',(npointG,ndim),dtype=xyz.dtype)
	dpoinO = file.create_dataset('order',(npointG,),dtype='i4')
	# Skip master if needed
	if ptable.has_master and MPI_RANK == 0: return None, None, None
	# Skip empty part
	if order.shape[0] == 0: return None, None, None
	# Get the position where the points should be stored
	inods,idx = np.unique(order,return_index=True)
	# Write dataset - points
	dxyz[inods,:] = xyz[idx,:]
	dpoinO[inods] = order[idx]
	return inods,idx,npointG

def h5_load_dset_size(file):
	'''
	Load only the number of points for the dataset
	'''
	# Crash if mesh is not present
	if not 'xyz' in file.keys():
		raiseError('Repartition is not possible without a the points!')
	# If the mesh is present read the size
	npoints = int(file['npoints'][0])
	point   = int(file['pointData'][0])
	return npoints, point

def h5_load_points(file,ptable,point):
	'''
	Load the mesh inside the HDF5 file
	'''
	if ptable.nodes is None or not point:
		# Warning! Repartition will only work if the input file is serial
		# i.e., it does not have any repeated nodes, otherwise it wont work
		istart, iend = ptable.partition_bounds(MPI_RANK,points=point)
		ptable.nodes = np.arange(istart,iend,dtype=np.int32)
	inods = ptable.nodes
	xyz   = np.array(file['xyz'][inods,:]) 
	order = np.array(file['order'][inods])
	# Return
	return xyz, order

def h5_create_variable_datasets(file,varDict,ptable,ipart=-1):
	'''
	Create the variable datasets inside an HDF5 file
	'''
	# Create group for variables
	group = file.create_group('VARIABLES_%d'%ipart if ipart >= 0 else 'VARIABLES')
	dsetDict = {}
	for var in varDict.keys():
		vargroup = group.create_group(var)
		nvars    = varDict[var]['value'].shape[0]
		dsetDict[var] = {
			'idim'  : vargroup.create_dataset('idim' ,(1,),dtype='i4'),
			'value' : vargroup.create_dataset('value',(nvars,),dtype=varDict[var]['value'].dtype),
		}
	return dsetDict

def h5_fill_variable_datasets(dsetDict,varDict):
	'''
	Fill in the variable datasets inside an HDF5 file
	'''
	for var in dsetDict.keys():
		# Fill dataset
		dsetDict[var]['idim'][:]  = varDict[var]['idim']
		dsetDict[var]['value'][:] = varDict[var]['value']

def h5_load_variables_single(file):
	'''
	Load the variables inside the HDF5 file
	'''
	varDict = {}
	for v in file['VARIABLES'].keys():
		vargroup = file['VARIABLES'][v]
		varDict[v] = {
			'idim'  : int(vargroup['idim'][0]),
			'value' : np.array(vargroup['value']),
		}
	# Return
	return varDict

def h5_load_variables_multi(file,npart):
	'''
	Load the variables inside the HDF5 file
	'''
	# Scan for variables in first partition and build variable dictionary
	varDict = {}
	for v in file['VARIABLES_0'].keys():
		vargroup = file['VARIABLES_0'][v]
		# Load point and dimensions
		idim = int(vargroup['idim'][0])
		# Now allocate output array
		value =  np.array(vargroup['value'])
		# Generate dictionary
		varDict[v] = {'idim':idim,'value':value}
	# Read variables per partition
	for ipart in range(1,npart):
		# Compute start and end of my partition in time
		vargroup = file['VARIABLES_%d'%ipart][v]
		value    =  np.array(vargroup['value'])
		varDict[v]['value'] = np.concatenate((varDict[v]['value'],value))
	# Return
	return varDict

def h5_create_field_datasets(file,fieldDict,ptable,ipart=-1):
	'''
	Create the variable datasets inside an HDF5 file
	'''
	# Create group for variables
	group = file.create_group('FIELDS_%d'%ipart if ipart >= 0 else 'FIELDS')
	dsetDict = {}
	for var in fieldDict.keys():
		vargroup = group.create_group(var)
		n     = mpi_reduce(fieldDict[var]['value'].shape[0] if not np.any(np.isnan(fieldDict[var]['value'])) else 0,op='sum',all=True)
		if ptable.has_master: n -= 1
		npoin = int(file['xyz'].shape[0])
		ndim  = n//npoin
		dims  = tuple([ndim*npoin] + [fieldDict[var]['value'].shape[ivar+1] for ivar in range(len(fieldDict[var]['value'].shape) - 1)])
		dsetDict[var] = {
			'ndim'  : vargroup.create_dataset('ndim' ,(1,),dtype='i4'),
			'nvar'  : vargroup.create_dataset('nvar' ,(1,),dtype='i4'),
			'vars'  : vargroup.create_dataset('vars' ,(len(fieldDict[var]['value'].shape) - 1,),dtype='i4'),
			'value' : vargroup.create_dataset('value',dims,dtype=fieldDict[var]['value'].dtype),
		}
	return dsetDict

def h5_fill_field_datasets(dsetDict,fieldDict,ptable,point,inods,idx):
	'''
	Fill in the variable datasets inside an HDF5 file
	'''
	# Skip master if needed
	if ptable.has_master and MPI_RANK == 0: return
	for var in dsetDict.keys():
		# Fill dataset
		dsetDict[var]['ndim'][:]  = fieldDict[var]['ndim']
		dsetDict[var]['nvar'][:]  = len(fieldDict[var]['value'].shape) - 1
		dsetDict[var]['vars'][:]  = fieldDict[var]['value'].shape[1:]
		# Compute start and end bounds for the variable
		if inods is None:
			istart, iend = ptable.partition_bounds(MPI_RANK,ndim=fieldDict[var]['ndim'],points=point)
			dsetDict[var]['value'][istart:iend,:] = fieldDict[var]['value']
		else:
			if fieldDict[var]['ndim'] > 1: raiseError('Cannot deal with multi-dimensional arrays in no partition mode!')
			dsetDict[var]['value'][inods,:] = fieldDict[var]['value'][idx,:]

def h5_load_fields_single(file,npoints,ptable,varDict,point):
	'''
	Load the fields inside the HDF5 file
	'''
	# Read variables
	fieldDict = {}
	for v in file['FIELDS'].keys():
		fieldgroup = file['FIELDS'][v]
		# Load point and dimensions
		ndim = int(fieldgroup['ndim'][0])
		dims = [ndim*npoints] + list(fieldgroup['vars'])
		# Now allocate output array
		value = np.zeros(dims,fieldgroup['value'])
		# Select which points to load
		if point:
			inods = ptable.partition_points(npoints,ndim=ndim)
		else:
			# Use the partition bounds to recover the array
			istart, iend = ptable.partition_bounds(MPI_RANK,ndim=ndim,points=False)
			inods = np.arange(istart,iend,dtype=np.int32)
		# Read the values
		value[:] = np.array(fieldgroup['value'][inods])
		# Generate dictionary
		fieldDict[v] = {'ndim':ndim,'value':value}
	# Return
	return fieldDict

def h5_load_fields_multi(file,npoints,ptable,varDict,point,npart):
	'''
	Load the fields inside the HDF5 file
	'''
	# Scan for variables in first partition and build variable dictionary
	fieldDict = {}
	for v in file['FIELDS_0'].keys():
		fieldgroup = file['FIELDS_0'][v]
		# Load point and dimensions
		ndim = int(fieldgroup['ndim'][0])
		dims = [ndim*npoints] + list(np.sum([file['FIELDS_%d'%ipart][v]['vars'] for ipart in range(npart)],axis=0))
		# Now allocate output array
		value = np.zeros(dims,fieldgroup['value'].dtype)	
		# Generate dictionary
		fieldDict[v] = {'ndim':ndim,'value':value}
	# Generate the partition size
	psize = [len(varDict[vv]['value'])//npart for vv in varDict.keys()]
	# Read variables per partition
	for ipart in range(npart):
		# Compute start and end of my partition in time
		pname  = 'FIELDS_%d'%ipart
		pstart = [ipart*p for p in psize]
		pend   = [(ipart+1)*p for p in psize]
		# Read the partition
		for v in file[pname].keys():
			fieldgroup = file[pname][v]
			# Load ndim
			ndim   = int(fieldgroup['ndim'][0])
			sliced = tuple([np.s_[:]] + [np.s_[i:j] for (i,j) in zip(pstart,pend)])
			# Select which points to load
			if point:
				inods = ptable.partition_points(npoints,ndim=ndim)
			else:
				# Use the partition bounds to recover the array
				istart, iend = ptable.partition_bounds(MPI_RANK,ndim=ndim,points=False)
				inods = np.arange(istart,iend,dtype=np.int32)
			# Read the values
			fieldDict[v]['value'][sliced] = np.array(fieldgroup['value'][inods])
	# Return
	return fieldDict


@cr('h5IO.save_dset')
def h5_save_dset(fname,xyz,varDict,fieldDict,ordering,point,ptable,mode='w',mpio=True,nopartition=False):
	'''
	Save a Dataset in HDF5
	'''
	if mpio and not MPI_SIZE == 1:
		h5_save_dset_mpio(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable,nopartition)
	else:
		h5_save_dset_serial(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable)

def h5_save_dset_serial(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable):
	'''
	Save a Dataset in HDF5 in serial mode
	'''
	# Open file for writing
	file = h5py.File(fname,mode)
	file.attrs['Version'] = PYLOM_H5_VERSION
	# Create dataset group
	group = file.create_group('DATASET')
	# Save points
	inods,idx,_ = h5_save_points(group,xyz,ordering,ptable,point)
	# Store the variables
	h5_fill_variable_datasets(h5_create_variable_datasets(group,varDict,ptable),varDict)
	# Store the fields
	h5_fill_field_datasets(h5_create_field_datasets(group,fieldDict,ptable),fieldDict,ptable,point,inods,idx)
	file.close()

def h5_save_dset_mpio(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable,nopartition):
	'''
	Save a Dataset in HDF5 in parallel mode
	'''
	# Open file
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	file.attrs['Version'] = PYLOM_H5_VERSION
	# Create dataset group
	group = file.create_group('DATASET')
	# Save points
	inods,idx,_ = h5_save_points(group,xyz,ordering,ptable,point) if not nopartition else h5_save_points_nopartition(group,xyz,ordering,ptable,point)
	# Store the variables
	h5_fill_variable_datasets(h5_create_variable_datasets(group,varDict,ptable),varDict)
	# Store the fields
	h5_fill_field_datasets(h5_create_field_datasets(group,fieldDict,ptable),fieldDict,ptable,point,inods,idx)
	file.close()


@cr('h5IO.append_dset')
def h5_append_dset(fname,xyz,varDict,fieldDict,ordering,point,ptable,mode='a',mpio=True,nopartition=False):
	'''
	Save a Dataset in HDF5
	'''
	if mpio and not MPI_SIZE == 1:
		h5_append_dset_mpio(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable,nopartition)
	else:
		h5_append_dset_serial(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable)

def h5_append_dset_serial(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable):
	'''
	Save a dataset in HDF5 in serial mode
	'''
	file = h5py.File(fname,mode)
	if not hasattr(h5_append_dset_serial,'ipart'):
		# Input file does not exist, we create it with the whole structure
		file.attrs['Version'] = PYLOM_H5_VERSION
		# Create dataset group
		group = file.create_group('DATASET')
		# Save points
		inods,idx,npoints = h5_save_points(group,xyz,ordering,ptable,point)
		# Start the partition counter
		h5_append_dset_serial.ipart   = 0
		h5_append_dset_serial.inods   = inods
		h5_append_dset_serial.idx     = idx
		h5_append_dset_serial.npoints = npoints
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Obtain from function
	group   = file['DATASET']
	ipart   = h5_append_dset_serial.ipart
	inods   = h5_append_dset_serial.inods
	idx     = h5_append_dset_serial.idx
	npoints = h5_append_dset_serial.npoints 
	# Store the variables
	h5_fill_variable_datasets(h5_create_variable_datasets(group,varDict,ptable,ipart=ipart),varDict)
	# Store the fields
	h5_fill_field_datasets(h5_create_field_datasets(group,fieldDict,ptable,ipart=ipart),fieldDict,ptable,point,inods,idx)
	# Increase the partition counter
	h5_append_dset_serial.ipart += 1
	file.close()

def h5_append_dset_mpio(fname,mode,xyz,varDict,fieldDict,ordering,point,ptable,nopartition):
	'''
	Save a dataset in HDF5 in parallel mode
	'''
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	if not hasattr(h5_append_dset_mpio,'ipart'):
		# Input file does not exist, we create it with the whole structure
		file.attrs['Version'] = PYLOM_H5_VERSION
		# Create dataset group
		group = file.create_group('DATASET')
		# Save points
		inods,idx,npoints = h5_save_points(group,xyz,ordering,ptable,point) if not nopartition else h5_save_points_nopartition(group,xyz,ordering,ptable,point)
		# Start the partition counter
		h5_append_dset_mpio.ipart   = 0
		h5_append_dset_mpio.inods   = inods
		h5_append_dset_mpio.idx     = idx
		h5_append_dset_mpio.npoints = npoints
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Obtain from function
	group   = file['DATASET']
	ipart   = h5_append_dset_mpio.ipart
	inods   = h5_append_dset_mpio.inods
	idx     = h5_append_dset_mpio.idx
	npoints = h5_append_dset_mpio.npoints 
	# Store the variables
	h5_fill_variable_datasets(h5_create_variable_datasets(group,varDict,ptable,ipart=ipart),varDict)
	# Store the fields
	h5_fill_field_datasets(h5_create_field_datasets(group,fieldDict,ptable,ipart=ipart),fieldDict,ptable,point,inods,idx)
	# Increase the partition counter
	h5_append_dset_mpio.ipart += 1
	file.close()


@cr('h5IO.load_dset')
def h5_load_dset(fname,ptable=None,mpio=True):
	'''
	Load a dataset in HDF5
	'''
	if mpio and not MPI_SIZE == 1:
		return h5_load_dset_mpio(fname,ptable)
	else:
		return h5_load_dset_serial(fname,ptable)

def h5_load_dset_serial(fname,ptable):
	'''
	Load a dataset in HDF5 in serial
	'''
	# Open file for writing
	file = h5py.File(fname,'r')
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Open the dataset group
	group  = file['DATASET']
	# Read dataset size
	npoints,point = h5_load_dset_size(group)
	# Are we reading for the same number of partitions?
	if ptable is None or not ptable.check_split():
		# Redo the partitions table
		ptable = PartitionTable.new(MPI_SIZE,npoints,npoints)
	# Read the points
	xyz, order = h5_load_points(group,ptable,point)
	# Figure out how many partitions we have
	npart = np.sum(['VAR' in key for key in group.keys()])
	# Read the variables
	varDict   = h5_load_variables_single(group) if npart == 1 else h5_load_variables_multi(group,npart)
	fieldDict = h5_load_fields_single(group,npoints,ptable,varDict,point) if npart == 1 else h5_load_fields_multi(group,npoints,ptable,varDict,point,npart)
	file.close()
	return xyz, order, point, ptable, varDict, fieldDict

def h5_load_dset_mpio(fname,ptable):
	'''
	Load a field in HDF5 in parallel
	'''
	# Open file for reading
	file = h5py.File(fname,'r',driver='mpio',comm=MPI_COMM)
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Open the dataset group
	group  = file['DATASET']
	# Read dataset size
	npoints,point = h5_load_dset_size(group)
	# Are we reading for the same number of partitions?
	if ptable is None or not ptable.check_split():
		# Redo the partitions table
		ptable = PartitionTable.new(MPI_SIZE,npoints,npoints)
	# Read the points
	xyz, order = h5_load_points(group,ptable,point)
	# Figure out how many partitions we have
	npoints = xyz.shape[0]
	npart   = np.sum(['VAR' in key for key in group.keys()])
	# Read the variables
	varDict   = h5_load_variables_single(group) if npart == 1 else h5_load_variables_multi(group,npart)
	fieldDict = h5_load_fields_single(group,npoints,ptable,varDict,point) if npart == 1 else h5_load_fields_multi(group,npoints,ptable,varDict,point,npart)
	file.close()
	return xyz, order, point, ptable, varDict, fieldDict


@cr('h5IO.save_mesh')
def h5_save_mesh(fname,mtype,xyz,conec,eltype,cellO,pointO,ptable,mode='w',mpio=True,nopartition=False):
	'''
	Save a Mesh in HDF5
	'''
	if mpio and not MPI_SIZE == 1:
		h5_save_mesh_mpio(fname,mode,mtype,xyz,conec,eltype,cellO,pointO,ptable,nopartition)
	else:
		h5_save_mesh_serial(fname,mode,mtype,xyz,conec,eltype,cellO,pointO,ptable)

def h5_save_mesh_serial(fname,mode,mtype,xyz,conec,eltype,cellO,pointO,ptable):
	'''
	Save a Mesh in HDF5 in serial mode
	'''
	# Open file for writing
	file = h5py.File(fname,mode)
	file.attrs['Version'] = PYLOM_H5_VERSION
	# Store partition table
	h5_save_partition(file,ptable)
	# Create dataset group
	group = file.create_group('MESH')
	# Save mesh
	h5_save_meshes(group,mtype,xyz,conec,eltype,cellO,pointO,ptable)
	file.close()

def h5_save_mesh_mpio(fname,mode,mtype,xyz,conec,eltype,cellO,pointO,ptable,nopartition):
	'''
	Save a dataset in HDF5 in parallel mode
	'''
	# Open file
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM)
	file.attrs['Version'] = PYLOM_H5_VERSION
	# Store partition table
	h5_save_partition(file,ptable)
	# Create dataset group
	group = file.create_group('MESH')
	# Save mesh
	h5_save_meshes(group,mtype,xyz,conec,eltype,cellO,pointO,ptable) if not nopartition else h5_save_meshes_nopartition(group,mtype,xyz,conec,eltype,cellO,pointO,ptable)
	file.close()


@cr('h5IO.load_mesh')
def h5_load_mesh(fname,mpio=True):
	'''
	Load a mesh in HDF5
	'''
	if mpio and not MPI_SIZE == 1:
		return h5_load_mesh_mpio(fname)
	else:
		return h5_load_mesh_serial(fname)

def h5_load_mesh_serial(fname):
	'''
	Load a mesh in HDF5 in serial
	'''
	# Open file for writing
	file = h5py.File(fname,'r')
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Read partition table
	ptable = h5_load_partition(file)
	repart = False
	# Are we reading for the same number of partitions?
	group = file['MESH']
	if not ptable.check_split():
		# Read the number of elements and points to compute
		# the new partition table
		npoints, ncells = h5_load_meshes_size(group)
		# Redo the partitions table
		ptable = PartitionTable.new(MPI_SIZE,ncells,npoints)
		repart = True
	# Read the mesh
	if not 'MESH' in file.keys(): raiseError('Mesh not present in dataset!')
	mtype,xyz,conec,eltype,cellO,pointO = h5_load_meshes(group,ptable,repart)
	# Close the file
	file.close()
	return mtype, xyz, conec, eltype, cellO, pointO, ptable

def h5_load_mesh_mpio(fname):
	'''
	Load a mesh in HDF5 in parallel
	'''
	# Open file for reading
	file = h5py.File(fname,'r',driver='mpio',comm=MPI_COMM)
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Read partition table
	ptable = h5_load_partition(file)
	repart = False
	# Are we reading for the same number of partitions?
	group = file['MESH']
	if not ptable.check_split():
		# Read the number of elements and points to compute
		# the new partition table
		npoints, ncells = h5_load_meshes_size(group)
		# Redo the partitions table
		ptable = PartitionTable.new(MPI_SIZE,ncells,npoints)
		repart = True
	# Read the mesh
	if not 'MESH' in file.keys(): raiseError('Mesh not present in dataset!')
	mtype,xyz,conec,eltype,cellO,pointO = h5_load_meshes(group,ptable,repart)
	# Close the file
	file.close()
	return mtype, xyz, conec, eltype, cellO, pointO, ptable


@cr('h5IO.save_POD')
def h5_save_POD(fname,U,S,V,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store POD variables into an HDF5 file.
	Can be appended to another HDF by setting the
	mode to 'a'. Then no partition table will be saved.
	'''
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,mode)
	# Store attributes and partition table
	if not mode == 'a':
		file.attrs['Version'] = PYLOM_H5_VERSION
		# Store partition table
		h5_save_partition(file,ptable)
	# Now create a POD group
	group = file.create_group('POD')
	# Create the datasets for U, S and V
	group.create_dataset('pointData',(1,),dtype='u1',data=pointData)
	group.create_dataset('n_variables',(1,),dtype='u1',data=nvars)
	Usize = (mpi_reduce(U.shape[0],op='sum',all=True),U.shape[1])
	dsetU = group.create_dataset('U',Usize,dtype=U.dtype)
	dsetS = group.create_dataset('S',S.shape,dtype=S.dtype)
	dsetV = group.create_dataset('V',V.shape,dtype=V.dtype)
	# Store S and U that are repeated across the ranks
	# So it is enough that one rank stores them
	if is_rank_or_serial(0):
		dsetS[:] = S
		dsetV[:] = V
	# Store U in parallel
	istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars,points=pointData)
	dsetU[istart:iend,:] = U
	file.close()

@cr('h5IO.load_POD')
def h5_load_POD(fname,vars,nmod,ptable=None):
	'''
	Load POD variables from an HDF5 file.
	'''
	file = h5py.File(fname,'r',driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,'r')
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Read the requested variables S, V
	varList = []
	if 'U' in vars:
		# Check if we need to read the partition table
		if ptable is None: ptable = h5_load_partition(file)
		# Read
		nvars = int(file['POD']['n_variables'][0])
		point = bool(file['POD']['pointData'][0])
		istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars,points=point)
		varList.append( np.array(file['POD']['U'][istart:iend,:nmod]) )
	if 'S' in vars: varList.append( np.array(file['POD']['S'][:]) )
	if 'V' in vars: varList.append( np.array(file['POD']['V'][:,:]) )
	# Return
	file.close()
	return varList


@cr('h5IO.save_DMD')
def h5_save_DMD(fname,muReal,muImag,Phi,bJov,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store DMD variables into an HDF5 file.
	Can be appended to another HDF by setting the
	mode to 'a'. Then no partition table will be saved.
	'''
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,mode)
	# Store attributes and partition table
	if not mode == 'a':
		file.attrs['Version'] = PYLOM_H5_VERSION
		# Store partition table
		h5_save_partition(file,ptable)
	# Now create a POD group
	group = file.create_group('DMD')
	# Create the datasets for U, S and V
	group.create_dataset('pointData',(1,),dtype='u1',data=pointData)
	group.create_dataset('n_variables',(1,),dtype='u1',data=nvars)
	Phisz = (mpi_reduce(Phi.shape[0],op='sum',all=True),Phi.shape[1])
	dsPhi = group.create_dataset('Phi',Phisz,dtype=Phi.dtype)
	dsMu  = group.create_dataset('Mu',(muReal.shape[0],2),dtype=muReal.dtype)
	dsJov = group.create_dataset('bJov',bJov.shape,dtype=bJov.dtype)
	# Store S and U that are repeated across the ranks
	# So it is enough that one rank stores them
	if is_rank_or_serial(0):
		dsMu[:,0] = muReal
		dsMu[:,1] = muImag
		dsJov[:]  = bJov
	# Store U in parallel
	istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars,points=pointData)
	dsPhi[istart:iend,:] = Phi
	file.close()

@cr('h5IO.load_DMD')
def h5_load_DMD(fname,vars,nmod,ptable=None):
	'''
	Load DMD variables from an HDF5 file.
	'''
	file = h5py.File(fname,'r',driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,'r')
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Read the requested variables S, V
	varList = []
	if 'Phi' in vars:
		# Check if we need to read the partition table
		if ptable is None: ptable = h5_load_partition(file)
		# Read
		nvars = int(file['DMD']['n_variables'][0])
		point = bool(file['DMD']['pointData'][0])
		istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars,point=point)
		varList.append( np.array(file['DMD']['Phi'][istart:iend,:nmod]) )
	if 'mu' in vars: 
		varList.append( np.array(file['DMD']['Mu'][:,0]) ) # Real
		varList.append( np.array(file['DMD']['Mu'][:,1]) ) # Imag
	if 'bJov' in vars: varList.append( np.array(file['DMD']['bJov'][:]) )
	# Return
	file.close()
	return varList

@cr('h5IO.save_VAE')
def h5_save_VAE(fname, kld, mse, val_loss, train_loss_avg, corrcoef, mode='w'):
	'''
	Store VAE results.
	'''
	file = h5py.File('%s.h5'%fname,mode,driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,mode)
	# Now create a VAE group
	group = file.create_group('VAE')
	# Create the datasets for U, S and V
	
	group.create_dataset('kld',(kld.shape[0],),dtype='u1',data=kld)
	group.create_dataset('mse',(mse.shape[0],),dtype='u1',data=mse)
	group.create_dataset('val_loss',(val_loss.shape[0],),dtype='u1',data=val_loss)
	group.create_dataset('train_loss_avg',(train_loss_avg.shape[0],),dtype='u1',data=train_loss_avg)
	group.create_dataset('correlation',(corrcoef.shape[0],),dtype='u1',data=corrcoef)
	file.close()

def h5_save_SPOD(fname,L,P,f,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store SPOD variables into an HDF5 file.
	Can be appended to another HDF by setting the
	mode to 'a'. Then no partition table will be saved.
	'''
	file = h5py.File(fname,mode,driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,mode)
	# Store attributes and partition table
	if not mode == 'a':
		file.attrs['Version'] = PYLOM_H5_VERSION
		# Store partition table
		h5_save_partition(file,ptable)
	# Get number of blocks
	nblocks = L.shape[1]
	# Now create a POD group
	group = file.create_group('SPOD')
	# Create the datasets for U, S and V
	group.create_dataset('pointData',(1,),dtype='u1',data=pointData)
	group.create_dataset('n_variables',(1,),dtype='u1',data=nvars)
	group.create_dataset('n_blocks',(1,),dtype='u1',data=nblocks)
	Psz = (mpi_reduce(P.shape[0],op='sum',all=True),P.shape[1])
	dsP = group.create_dataset('P',Psz,dtype=P.dtype)
	dsL = group.create_dataset('L',L.shape,dtype=L.dtype)
	dsf = group.create_dataset('f',f.shape,dtype=f.dtype)
	# Store L and f that are repeated across the ranks (nblocks,nfreq)
	# So it is enough that one rank stores them
	if is_rank_or_serial(0):
		dsL[:,:] = L
		dsf[:]   = f
	# Store P in parallel (nblocks*nvars*npoints,nfreq)
	istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars*nblocks,points=pointData)
	dsP[istart:iend,:] = P
	file.close()

def h5_load_SPOD(fname,vars,nmod,ptable=None):
	'''
	Load SPOD variables from an HDF5 file.
	'''
	file = h5py.File(fname,'r',driver='mpio',comm=MPI_COMM) if not MPI_SIZE == 1 else h5py.File(fname,'r')
	# Check the file version
	version = tuple(file.attrs['Version'])
	if not version == PYLOM_H5_VERSION:
		raiseError('File version <%s> not matching the tool version <%s>!'%(str(file.attrs['Version']),str(PYLOM_H5_VERSION)))
	# Read the requested variables S, V
	varList = []
	if 'P' in vars:
		# Check if we need to read the partition table
		if ptable is None: ptable = h5_load_partition(file)
		# Read
		nvars   = int(file['SPOD']['n_variables'][0])
		nblocks = int(file['SPOD']['n_blocks'][0])
		point   = bool(file['SPOD']['pointData'][0])
		istart, iend = ptable.partition_bounds(MPI_RANK,ndim=nvars*nblocks,point=point)
		varList.append( np.array(file['SPOD']['P'][istart:iend,:nmod]) )
	if 'L' in vars: 
		varList.append( np.array(file['SPOD']['L'][:,:]) )
	if 'f' in vars: 
		varList.append( np.array(file['SPOD']['f'][:]) )
	# Return
	file.close()
	return varList



#!/usr/bin/env python
#
# pyLOM, utils.
#
# Parallel MPI routines
#
# Last rev: 14/02/2025
from __future__ import print_function, division

import mpi4py, numpy as np
mpi4py.rc.recv_mprobe = False
from mpi4py import MPI

from .parall import split
from .nvtxp  import nvtxp

MPI_COMM = MPI.COMM_WORLD
MPI_RANK = MPI_COMM.Get_rank()
MPI_SIZE = MPI_COMM.Get_size()

MPI_RDONLY = MPI.MODE_RDONLY
MPI_WRONLY = MPI.MODE_WRONLY
MPI_CREATE = MPI.MODE_CREATE


# Expose functions from MPI library
mpi_create_op = MPI.Op.Create
mpi_wtime     = MPI.Wtime
mpi_file_open = MPI.File.Open

mpi_nanmin = mpi_create_op(lambda v1,v2,dtype : np.nanmin([v1,v2]),commute=True)
mpi_nanmax = mpi_create_op(lambda v1,v2,dtype : np.nanmax([v1,v2]),commute=True)
mpi_nansum = mpi_create_op(lambda v1,v2,dtype : np.nansum([v1,v2]),commute=True)


def mpi_barrier():
	'''
	Implements the barrier
	'''
	MPI_COMM.Barrier()


@nvtxp('mpi_send',color='red')
def mpi_send(f,dest,tag=0):
	'''
	Implements the send operation
	'''
	MPI_COMM.send(f,dest,tag=tag)


@nvtxp('mpi_recv',color='red')
def mpi_recv(**kwargs):
	'''
	Implements the recieve operation
	'''
	return MPI_COMM.recv(**kwargs)


@nvtxp('mpi_sendrecv',color='red')
def mpi_sendrecv(buff,**kwargs):
	'''
	Implements the sendrecv operation
	'''
	return MPI_COMM.sendrecv(buff,**kwargs)


@nvtxp('mpi_scatter',color='red')
def mpi_scatter(sendbuff,root=0,do_split=False):
	'''
	Send an array among the processors and split
	if necessary.
	'''
	if MPI_SIZE > 1:
		return MPI_COMM.scatter(split(sendbuff,root=root),root=root) if do_split else MPI_COMM.scatter(sendbuff,root=root)
	return sendbuff


@nvtxp('mpi_gather',color='red')
def mpi_gather(sendbuff,root=0,all=False):
	'''
	Gather an array from all the processors.
	'''
	if MPI_SIZE > 1:
		if not isinstance(sendbuff,np.ndarray) and not isinstance(sendbuff,list): sendbuff = [sendbuff]
		if all:
			out = MPI_COMM.allgather(sendbuff)
			return np.concatenate(out,axis=0)
		else:
			out = MPI_COMM.gather(sendbuff,root=root)
			return np.concatenate(out,axis=0) if MPI_RANK == root else None
	return sendbuff


@nvtxp('mpi_reduce',color='red')
def mpi_reduce(sendbuff,root=0,op='sum',all=False):
	'''
	Reduce an array from all the processors.
	'''
	if MPI_SIZE > 1:
		if isinstance(op,str):
			if 'sum'    in op: opf = MPI.SUM
			if 'max'    in op: opf = MPI.MAX
			if 'min'    in op: opf = MPI.MIN
			if 'nanmin' in op: opf = mpi_nanmin
			if 'nanmax' in op: opf = mpi_nanmax
			if 'nansum' in op: opf = mpi_nansum
		else:
			opf = op
		if all:
			return MPI_COMM.allreduce(sendbuff,op=opf)
		else:
			out = MPI_COMM.reduce(sendbuff,op=opf,root=root)
			return out if root == MPI_RANK else sendbuff
	else:
		return sendbuff


@nvtxp('mpi_bcast',color='red')
def mpi_bcast(sendbuff,root=0):
	'''
	Implements the broadcast operation
	'''
	return MPI_COMM.bcast(sendbuff,root=root)



#!/usr/bin/env python
#
# PYLOM Testsuite Isomap
#
# Last revision: 11/01/2025
from __future__ import print_function, division

import sys, json
import pyLOM

## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"'))

## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X = d.X(*VARIABLES)
t = d.get_variable('time')

# Define the number of neighbours
K = PARAMS['K']

## Run Isomap
Y,R,_ = pyLOM.MANIFOLD.isomap(X,2,K)
## Testsuite output
pyLOM.pprint(0,'TSUITE R = %.10f'%R)
values = ''
for val in Y[0,:]:
    values += str(abs(val)) + ' '
pyLOM.pprint(0,'TSUITE X = ',values)
values = ''
for val in Y[1,:]:
    values += str(abs(val)) + ' '
pyLOM.pprint(0,'TSUITE Y = ',values)

## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# Example of GPOD - Reconstruct the Gappy Snapshot Matrix.
#
# Last revision: 26/02/2025
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt
import pyLOM
import pyLOM.utils

pyLOM.gpu_device(gpu_per_node=4)


## Parameters
DATAFILE = "./DATA/CYLINDER.h5"
VARIABLE = "VELOX"


## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE, ptable=m.partition_table).to_gpu([VARIABLE])
X = d[VARIABLE]
t = d.get_variable('time')
xyzc = m.xyzc


## Define percentage for missing data
missing_percentage = 0.15  # 15% missing data
# Generate random mask
np.random.seed(5)
random_mask = np.random.choice(
    [0, 1], size=X.shape, p=[missing_percentage, 1 - missing_percentage]
)
# Create incomplete snapshot matrix
incomplete_snapshot = X * pyLOM.utils.cpu_to_gpu(random_mask)

# Database reconstruction
num_iter = 50

# GappyPOD model setup
gappy_model_recons = pyLOM.GPOD.GappyPOD(
    centered=False,
    apply_truncation=True,
    truncation_param=-0.98,
    reconstruction_method="ridge",
    ridge_lambda=0.1,
)


## Reconstruct database
X_recons, eig_spec_iter, c_e = gappy_model_recons.reconstruct_full_set(
    incomplete_snapshot,
    num_iter,
)


## Compute metrics
mae_recons  = pyLOM.math.MAE(X, X_recons)
rmse_recons = pyLOM.math.RMSE(X, X_recons, relative=False)
r2_recons   = pyLOM.math.r2(X, X_recons)
print(f"MAE_database = {mae_recons}\nRMSE_database = {rmse_recons}\nR2_database = {r2_recons}")


## Eigenvalue spectrum and cumulative energy for original data
PSI, S, V = pyLOM.POD.run(X, remove_mean=True)
c_e_exact = S.cumsum() / S.sum()
S2 = S*S
eig_spec_exact = S2 / S2.sum()
mode_index = np.arange(1, len(c_e_exact) + 1)


## Add to dataset and send to CPU
d.add_field('VELXR',1,X_recons)
d.to_cpu(['VELOX','VELXR'])
c_e            = pyLOM.utils.gpu_to_cpu(c_e)
c_e_exact      = pyLOM.utils.gpu_to_cpu(c_e_exact)
eig_spec_exact = pyLOM.utils.gpu_to_cpu(eig_spec_exact)
eig_spec_iter  = pyLOM.utils.gpu_to_cpu(eig_spec_iter)

# Plot Eigenvalue spectrum
fig, ax = plt.subplots(figsize=(8, 9))
ax.scatter(mode_index[:-1], eig_spec_exact[:-1], c="black", label="Exact")
for it, label in zip(
    [0, 10, 25, num_iter - 1], ["1 iter", "10 iter", "25 iter", f"{num_iter} iter"]
):
    ax.plot(mode_index[:-1], eig_spec_iter[:, it][:-1], label=label)
ax.set_yscale("log")
ax.set_title("Ridge Gappy POD")
ax.legend()
ax.set_xlabel("Eigenvalue")
ax.set_ylabel("Eigenvalue spectrum")

# Plot Cumulative energy
fig, ax = plt.subplots(figsize=(8, 9))
ax.scatter(mode_index[:-1], c_e_exact[:-1], c="black", label="Exact")
for it, label in zip(
    [0, 10, 25, num_iter - 1], ["1 iter", "10 iter", "25 iter", f"{num_iter} iter"]
):
    ax.plot(mode_index[:-1], c_e[:, it][:-1], label=label)
ax.axhline(y=0.9, color="black", linestyle="--", label="90% Energy")
ax.set_title("Ridge Gappy POD")
ax.legend()
ax.set(xlabel="Eigenvalue", ylabel="Cumulative energy")

# Dump to ParaView
pyLOM.io.pv_writer(m,d,'flow',basedir='out/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=['VELOX','VELXR'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d,vars=['VELXR'],instant=0,component=0,cmap='jet',cpos='xy')


pyLOM.cr_info()
plt.show()


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Plotting utilities.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from ..vmmath import vector_norm
from ..utils  import cr_nvtx as cr, gpu_to_cpu, raiseWarning


def show_plots():
	'''
	Wrapper to matplotlib.pyplot.show()
	'''
	plt.show()

def close_plots():
	'''
	Wrapper to matplotlib.pyplot.close()
	'''
	plt.close()


def plotFieldStruct2D(ax,nx,ny,ndim,xyz,field,dim,cmap,clear=False):
	'''
	Plot a 2D point or cell centered field on an structured mesh
	'''
	# Clear the axis if needed
	if clear: ax.clear()
	# Obtain the colormap if needed
	if cmap is None: cmap = plt.get_cmap('coolwarm',256)
	# Obtain X, Y matrices
	X = xyz[:,0].reshape((nx,ny),order='c').T
	Y = xyz[:,1].reshape((nx,ny),order='c').T
	# Obtain data matrix
	Z = field.reshape((nx,ny,ndim),order='c').T if ndim > 1 else field.reshape((nx,ny),order='c').T
	levels = np.linspace(-1e-2, 1e-2, 11)
	return ax.contourf(X,Y,Z[ndim,:,:] if dim >= 0 else np.linalg.norm(Z,axis=0) if ndim > 1 else Z,cmap=cmap)


def plotResidual(S,fig=None,ax=None):
	'''
	Given the S matrix as a vector, plot
	the accumulated residual.
	'''
	S = gpu_to_cpu(S)
	# Build figure and axes
	if fig is None:
		fig = plt.figure(figsize=(8,6),dpi=100)
	if ax is None:
		ax = fig.add_subplot(1,1,1)
	# Compute accumulated residual
	accumulative_S  = np.array([vector_norm(S,ii) for ii in range(S.shape[0])],np.double)
	accumulative_S /= accumulative_S[0]
	# Plot accumulated residual
	ax.semilogy(np.arange(1,S.shape[0]+1),accumulative_S,'bo')
	# Set labels
	ax.set_ylabel(r'$\varepsilon_1$')
	ax.set_xlabel(r'Truncation size')
	ax.set_title(r'Tolerance')
	# Return
	return fig, ax


try:
	import pyvista as pv

	def _cells_and_offsets(conec):
		'''
		Build the offsets and cells array to create am
		UnstructuredGrid
		'''
		# Compute points per cell
		ppcell = np.sum(conec >= 0,axis=1)
		# Compute cells for pyVista, with the number of points per cell on front
		cells = np.c_[ppcell,conec]
		# Now we get rid of any -1 entries for mixed meshes
		cellsf = cells.flatten('c')
		cellsf = cellsf[cellsf>=0]
		# Now build the offsets vector
		offset = np.zeros((ppcell.shape[0]+1,),np.int32)
		offset[1:] = np.cumsum(ppcell)
		return cellsf, offset

	@cr('plots.pyvista_snap')
	def plotSnapshot(mesh,dset,vars=[],idim=0,instant=0,**kwargs):
		'''
		Plot using pyVista
		'''
		# First create the unstructured grid
		cells, offsets = _cells_and_offsets(mesh.connectivity)
		# Create the unstructured grid
		ugrid =  pv.UnstructuredGrid(offsets,cells,mesh.eltype2VTK,mesh.xyz) if pv.vtk_version_info < (9,) else pv.UnstructuredGrid(cells,mesh.eltype2VTK,mesh.xyz)
		# Load the variables inside the unstructured grid
		for v in vars:
			info   = dset.info(v)
			sliced = tuple([np.s_[:]] + [0 if i != idim else instant for i in range(len(dset[v].shape)-1)])
			if info['point']:
				ugrid.point_data[v] = mesh.reshape_var(dset[v][sliced],info)
			else:
				ugrid.cell_data[v]  = mesh.reshape_var(dset[v][sliced],info)
		# Launch plot
		return ugrid.plot(**kwargs)

	@cr('plots.pyvista_layout')
	def plotLayout(mesh,dset,nrows,ncols,imode,vars=[],cmap='jet',title='',off_screen=False,**kwargs):
		'''
		Plot using pyVista
		'''
		# First create the unstructured grid
		cells, offsets = _cells_and_offsets(mesh.connectivity)
		# Create the unstructured grid
		ugrid =  pv.UnstructuredGrid(offsets,cells,mesh.eltype2VTK,mesh.xyz) if pv.vtk_version_info < (9,) else pv.UnstructuredGrid(cells,mesh.eltype2VTK,mesh.xyz)
		# Load the variables inside the unstructured grid
		plotter = pv.Plotter(shape=(nrows,ncols),off_screen=off_screen)
		irow, icol = 0, 0
		for ivar, v in enumerate(vars):
			# Add variable to mesh
			info = dset.info(v)
			if info['point']:
				ugrid.point_data[v] = mesh.reshape_var(dset[v],info)
			else:
				ugrid.cell_data[v]  = mesh.reshape_var(dset[v],info)
			# Plot
			plotter.subplot(irow,icol)
			if ivar == 0: plotter.add_text(title)
			plotter.add_mesh(ugrid,scalars=v,cmap=cmap,component=imode,copy_mesh=True)
			# Check irow, icol bounds
			icol += 1
			if icol == ncols:
				icol  = 0
				irow += 1
		# Launch plot
		return plotter.show(**kwargs)

except:
	def plotSnapshot(mesh,dset,vars=[],idim=0,instant=0,**kwargs):
		'''
		Plot using pyVista
		'''
		raiseWarning('Import - Problems loading pyVista!',False)

	def plotLayout(mesh,dset,nrows,ncols,imode,vars=[],cmap='jet',title='',off_screen=False,**kwargs):
		'''
		Plot using pyVista
		'''
		raiseWarning('Import - Problems loading pyVista!',False)



#!/usr/bin/env python
#
# Example of DMD.
#
# Last revision: 27/01/2023
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import numpy as np
import pyLOM

pyLOM.gpu_device(gpu_per_node=4)


## Parameters
DATAFILE = './DATA/CYLINDER.h5'
VARIABLE = 'VELOX'


## Data loading
m  = pyLOM.Mesh.load(DATAFILE)
d  = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table).to_gpu([VARIABLE])
X  = d[VARIABLE]
t  = d.get_variable('time')
dt = t[1] - t[0]


## Run DMD
Y = X[:,:100].copy() # Grab the first 100 and run DMD
muReal,muImag,Phi,bJov = pyLOM.DMD.run(Y,1e-6,remove_mean=False)
# Compute frequency and damping ratio of the modes
delta, omega = pyLOM.DMD.frequency_damping(muReal,muImag,dt)
pyLOM.DMD.save('results.h5',muReal,muImag,Phi,bJov,d.partition_table,nvars=1,pointData=d.point)
# Reconstruction according to Jovanovic 2014
# Last 50 snapshots are predicted
X_DMD = pyLOM.DMD.reconstruction_jovanovic(Phi,muReal,muImag,t,bJov)
rmse  = pyLOM.math.RMSE(X_DMD.copy(),X.copy())
pyLOM.pprint(0,'RMSE = %e'%rmse)


## DMD plots
if pyLOM.utils.is_rank_or_serial(0):
	pyLOM.DMD.ritzSpectrum(muReal,muImag)               # Ritz Spectrum
	pyLOM.DMD.amplitudeFrequency(omega,bJov,norm=False) # Amplitude vs frequency
	pyLOM.DMD.dampingFrequency(omega,delta)             # Damping ratio vs frequency


## Dump to ParaView
# Spatial modes
d.add_field('U_MODES_REAL',6,pyLOM.DMD.extract_modes(Phi,1,len(d),real=True,modes=[1,4,6,2,5,3]))
d.add_field('U_MODES_IMAG',6,pyLOM.DMD.extract_modes(Phi,1,len(d),real=False,modes=[1,4,6,2,5,3]))
pyLOM.io.pv_writer(m,d.to_cpu(['U_MODES_REAL','U_MODES_IMAG']),'modes',basedir='out/modes',instants=[0],times=[0.],vars=['U_MODES_REAL','U_MODES_IMAG'],fmt='vtkh5')
pyLOM.DMD.plotMode(Phi,omega,m,d,1,pointData=False,modes=[1,2,3,4,5,6,7],cpos='xy')

# Temporal evolution
d.add_field('VELXR',1,X_DMD)
pyLOM.io.pv_writer(m,d.to_cpu(['VELOX','VELXR']),'flow',basedir='out/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=['VELOX','VELXR'],fmt='vtkh5')
pyLOM.DMD.plotSnapshot(m,d.to_cpu(['VELXR']),vars=['VELXR'],instant=0,cmap='jet',cpos='xy')


## Show and print timings
pyLOM.cr_info()
pyLOM.show_plots()


#!/usr/bin/env python
#
# pyLOM, mesh.
#
# Mesh class, to organize data.
#
# Last rev: 26/01/2023
from __future__ import print_function, division

import os, numpy as np

from .       import inp_out as io
from .vmmath import cellCenters, normals
from .utils  import cr_nvtx as cr, mem, raiseError, mpi_reduce


ALYA2ELTYP = {
	# Linear cells
	-1 :  0, # Empty cell
	 3 :  1, # Line element
	10 :  2, # Triangular cell
	12 :  3, # Quadrangular cell
	30 :  4, # Tetrahedral cell
	37 :  5, # Hexahedron
	34 :  6, # Linear prism
	32 :  7, # Pyramid
	# Quadratic, isoparametric cells
	39 : 15, # HEX27
	# Lagrangian cells
	40 : 25, # HEX64
}

ELTYPE2VTK = {
	 0 : 0,  # Empty cell
	 2 : 5,  # Triangular cell
	 3 : 9,  # Quadrangular cell
	 4 : 10, # Tetrahedral cell
	 6 : 13, # Linear prism
	 7 : 14, # Pyramid
	 5 : 12, # Hexahedron
	25 : 72, # Lagrangian Hexahedron
}

ELTYPE2ENSI = {
	2 : 'tria3',  # Triangular cell
	3 : 'quad4',  # Quadrangular cell
	4 : 'tetra4', # Tetrahedral cell
	6 : 'penta6', # Linear prism
	5 : 'hexa8',  # Hexahedron
}

MTYPE2ID = {
	'STRUCT2D' : 1,
	'STRUCT3D' : 2,
	'UNSTRUCT' : 3,
}

ID2MTYPE = {
	1 : 'STRUCT2D',
	2 : 'STRUCT3D',
	3 : 'UNSTRUCT',
}

class Mesh(object):
	'''
	The Mesh class wraps the mesh details of the case.
	'''
	def __init__(self,mtype,xyz,connectivity,eltype,cellOrder,pointOrder,ptable):
		'''
		Class constructor
		'''
		self._type   = mtype
		self._xyz    = xyz
		self._xyzc   = None
		self._normal = None
		self._conec  = connectivity
		self._eltype = eltype
		self._cellO  = cellOrder
		self._pointO = pointOrder
		self._ptable = ptable

	def __str__(self):
		'''
		String representation
		'''
		s   = 'Mesh (%s) of %d nodes and %d elements:\n' % (self.type,self.npoints,self.ncells)
		s  += '  > xyz  - max = ' + str(np.nanmax(self._xyz,axis=0)) + ', min = ' + str(np.nanmin(self._xyz,axis=0)) + '\n'
		return s

	def find_point(self,xyz):
		'''
		Return all the points where self._xyz == xyz
		'''
		return np.where(np.all(self._xyz == xyz,axis=1))[0]

	def find_cell(self,eltype):
		'''
		Return all the elements where self._elemList == elem
		'''
		return np.where(np.all(self._eltype == eltype))[0]

	def find_point_in_cell(self,inode):
		'''
		Return all the elements where the node is
		'''
		return np.where(np.any(np.isin(self._conec,inode),axis=1))[0]

	def size(self,pointData):
		'''
		Return the size accoding to the type of data
		'''
		return self.npoints if pointData else self.ncells

	@cr('Mesh.cellcenters')
	def cellcenters(self):
		'''
		Computes and returns the cell centers
		'''
		if self.type == 'STRUCT2D':
			# Recover unique X, Y coordinates
			x = np.unique(self.x)
			y = np.unique(self.y)
			# Compute cell centers
			xc = x[:-1] + np.diff(x)/2.
			yc = y[:-1] + np.diff(y)/2.
			# Build xyzc
			xx, yy    = np.meshgrid(xc,yc,indexing='ij')
			xyzc      = np.zeros((self.ncells,2),dtype=np.double)
			xyzc[:,0] = xx.reshape((self.ncells,),order='C')
			xyzc[:,1] = yy.reshape((self.ncells,),order='C')
		# Connectivity for a 3D mesh
		if self.type == 'STRUCT3D':
			# Recover unique X, Y, Z coordinates
			x = np.unique(self.x)
			y = np.unique(self.y)
			z = np.unique(self.z)
			# Compute cell centers
			xc = x[:-1] + np.diff(x)/2.
			yc = y[:-1] + np.diff(y)/2.
			zc = z[:-1] + np.diff(z)/2.
			# Build xyzc
			xx, yy, zz = np.meshgrid(xc,yc,zc,indexing='ij')
			xyzc       = np.zeros((self.ncells,3),dtype=np.double)
			xyzc[:,0]  = xx.reshape((self.ncells,),order='C')
			xyzc[:,1]  = yy.reshape((self.ncells,),order='C')		
			xyzc[:,2]  = zz.reshape((self.ncells,),order='C')		
		# Connectivity for a unstructured mesh
		if self.type == 'UNSTRUCT':
			xyzc = cellCenters(self._xyz,self._conec)
		return xyzc

	@cr('Mesh.reshape')
	def reshape_var(self,var,info):
		'''
		Reshape a variable according to the mesh
		'''
		# Obtain number of points from the mesh
		npoints = self.size(info['point'])
		# Only reshape the variable if ndim > 1
		out = np.ascontiguousarray(var.reshape((npoints,info['ndim']),order='C') if info['ndim'] > 1 else var)
		# Build 3D vector in case of 2D array
		if self.type == 'STRUCT2D' and info['ndim'] == 2:
			out = np.hstack((out,np.zeros((npoints,1))))
		return out

	@cr('Mesh.save')
	def save(self,fname,**kwargs):
		'''
		Store the mesh in various formats.
		'''
		# Guess format from extension
		fmt = os.path.splitext(fname)[1][1:] # skip the .
		# Pickle format
		if fmt.lower() == 'pkl': 
			io.pkl_save(fname,self)
		# H5 format
		if fmt.lower() == 'h5':
			# Set default parameters
			if not 'mode' in kwargs.keys():        kwargs['mode']        = 'w' if not os.path.exists(fname) else 'a'
			if not 'mpio' in kwargs.keys():        kwargs['mpio']        = True
			if not 'nopartition' in kwargs.keys(): kwargs['nopartition'] = False
			# Save
			io.h5_save_mesh(fname,self.type,self.xyz,self.connectivity,self.eltype,self.cellOrder,self.pointOrder,self.partition_table,**kwargs)

	@classmethod
	@cr('Mesh.load')
	def load(cls,fname,**kwargs):
		'''
		Load a mesh from various formats
		'''
		# Guess format from extension
		fmt = os.path.splitext(fname)[1][1:] # skip the .
		# Pickle format
		if fmt.lower() == 'pkl': 
			return io.pkl_load(fname)
		# H5 format
		if fmt.lower() == 'h5':
			if not 'mpio' in kwargs.keys(): kwargs['mpio'] = True
			mtype, xyz, conec, eltype, cellO, pointO, ptable = io.h5_load_mesh(fname,**kwargs)
			return cls(mtype,xyz,conec,eltype,cellO,pointO,ptable)
		raiseError('Cannot load file <%s>!'%fname)

	@classmethod
	@cr('Mesh.new_struct2D')
	def new_struct2D(cls,nx,ny,x,y,dimsx,dimsy,ptable=None):
		xyz    = _struct2d_compute_xyz(nx,ny,x,y,dimsx,dimsy)
		conec  = _struct2d_compute_conec(nx,ny,xyz)
		eltype = 3*np.ones(((nx-1)*(ny-1),),np.uint8)
		cellO  = np.arange((nx-1)*(ny-1),dtype=np.int32)
		pointO = np.arange(nx*ny,dtype=np.int32)
		return cls('STRUCT2D',xyz,conec,eltype,cellO,pointO,ptable)

	@classmethod
	@cr('Mesh.new_struct3D')
	def new_struct3D(cls,nx,ny,nz,x,y,z,dimsx,dimsy,dimsz,ptable=None):
		xyz    = _struct3d_compute_xyz(nx,ny,nz,x,y,z,dimsx,dimsy,dimsz)
		conec  = _struct3d_compute_conec(nx,ny,nz,xyz)
		eltype = 5*np.ones(((nx-1)*(ny-1)*(nz-1),),np.uint8)
		cellO  = np.arange((nx-1)*(ny-1)*(nz-1),dtype=np.int32)
		pointO = np.arange(nx*ny*nz,dtype=np.int32)
		return cls('STRUCT3D',xyz,conec,eltype,cellO,pointO,ptable)

	@classmethod
	@cr('Mesh.from_pyQvarsi')
	def from_pyQvarsi(cls,mesh,ptable=None,sod=False):
		'''
		Create the mesh structure from a pyQvarsi mesh structure
		'''
		eltype = np.array([ALYA2ELTYP[t] for t in mesh.eltype_linear],np.uint8)
		return cls('UNSTRUCT',mesh.xyz,mesh.connectivity_vtk if sod else mesh.connectivity,eltype,mesh.leinv_linear,mesh.lninv,ptable)

	@property
	def type(self):
		return self._type
	@property
	def npoints(self):
		return self._xyz.shape[0]
	@property
	def npointsG(self):
		return mpi_reduce(self.npoints,op='sum',all=True)
	@property
	def npointsG2(self):
		if self.pointOrder.shape[0] > 0:
			npoints = self.pointOrder.max()
		else:
			npoints = 0
		return mpi_reduce(npoints,op='max',all=True) + 1
	@property
	def ndim(self):
		return self._xyz.shape[1]
	@property
	def ncells(self):
		return self._eltype.shape[0]
	@property
	def ncellsG(self):
		return mpi_reduce(self.ncells,op='sum',all=True)
	@property
	def ncellsG2(self):
		if self.cellOrder.shape[0] > 0:
			ncells = self.cellOrder.max()
		else:
			ncells = 0
		return mpi_reduce(ncells,op='max',all=True) + 1
	@property
	def nnodcell(self):
		return self._conec.shape[1]

	@property
	def xyz(self):
		return self._xyz
	@property
	def x(self):
		return self._xyz[:,0]
	@property
	def y(self):
		return self._xyz[:,1]
	@property
	def z(self):
		return self._xyz[:,2]
	@property
	def xyzc(self):
		if self._xyzc is None: self._xyzc = self.cellcenters()
		return self._xyzc
	@property
	def normal(self):
		if self._normal is None: self._normal = normals(self._xyz,self._conec)
		return self._normal

	@property
	def connectivity(self):
		return self._conec
	@property
	def cellOrder(self):
		return self._cellO
	@property
	def pointOrder(self):
		return self._pointO

	@property
	def partition_table(self):
		return self._ptable
	@partition_table.setter
	def partition_table(self,value):
		self._ptable = value

	@property
	def eltype(self):
		return self._eltype
	@property
	def eltype2VTK(self):
		return np.array([ELTYPE2VTK[t] for t in self._eltype],np.uint8)
	@property
	def eltype2ENSI(self):
		return ELTYPE2ENSI[self._eltype[0]]


def _struct2d_compute_xyz(nx,ny,x,y,dimsx,dimsy):
	'''
	Compute points for a 2D structured mesh
	'''
	if x is None:
		dx = (dimsx[1] - dimsx[0])/(nx - 1.)
		x  = dx*np.arange(nx) + dimsx[0]
	if y is None:
		dy = (dimsy[1] - dimsy[0])/(ny - 1.)
		y  = dy*np.arange(ny) + dimsy[0]
	xx, yy = np.meshgrid(x,y,indexing='ij')
	xy = np.zeros((nx*ny,3),dtype=np.double)
	xy[:,0] = xx.reshape((nx*ny,),order='C')
	xy[:,1] = yy.reshape((nx*ny,),order='C')
	return xy

def _struct2d_compute_conec(nx,ny,xyz):
	'''
	Compute connectivity for a 2D structured mesh
	'''
	# Obtain the ids
	idx  = np.lexsort((xyz[:,1],xyz[:,0]))
	idx2 = idx.reshape((nx,ny))
	# Create connectivity array
	conec = np.zeros(((nx-1)*(ny-1),4),dtype=np.int32)
	conec[:,0] = idx2[:-1,:-1].ravel()
	conec[:,1] = idx2[:-1,1:].ravel()
	conec[:,2] = idx2[1:,1:].ravel()
	conec[:,3] = idx2[1:,:-1].ravel()
	return conec


def _struct3d_compute_xyz(nx,ny,nz,x,y,z,dimsx,dimsy,dimsz):
	'''
	Compute points for a 2D structured mesh
	'''
	if x is None:
		dx = (dimsx[1] - dimsx[0])/(nx - 1.)
		x  = dx*np.arange(nx) + dimsx[0]
	if y is None:
		dy = (dimsy[1] - dimsy[0])/(ny - 1.)
		y  = dy*np.arange(ny) + dimsy[0]
	if z is None:
		dz = (dimsz[1] - dimsz[0])/(nz - 1.)
		z  = dz*np.arange(nz) + dimsz[0]
	xx, yy, zz = np.meshgrid(x,y,z,indexing='ij')
	xyz = np.zeros((nx*ny*nz,3),dtype=np.double)
	xyz[:,0] = xx.reshape((nx*ny*nz,),order='C')
	xyz[:,1] = yy.reshape((nx*ny*nz,),order='C')
	xyz[:,2] = zz.reshape((nx*ny*nz,),order='C')
	return xyz

def _struct3d_compute_conec(nx,ny,nz,xyz):
	'''
	Compute connectivity for a 2D structured mesh
	'''
	# Obtain the ids
	idx  = np.lexsort((xyz[:,2],xyz[:,1],xyz[:,0]))
	idx2 = idx.reshape((nx,ny,nz))
	# Create connectivity array
	conec = np.zeros(((nx-1)*(ny-1)*(nz-1),8),dtype=np.int32)
	conec[:,0] = idx2[:-1,:-1,:-1].ravel()
	conec[:,1] = idx2[:-1,:-1,1:].ravel()
	conec[:,2] = idx2[:-1,1:,1:].ravel()
	conec[:,3] = idx2[:-1,1:,:-1].ravel()
	conec[:,4] = idx2[1:,:-1,:-1].ravel()
	conec[:,5] = idx2[1:,:-1,1:].ravel()
	conec[:,6] = idx2[1:,1:,1:].ravel()
	conec[:,7] = idx2[1:,1:,:-1].ravel()
	return conec



#!/usr/bin/env python
#
# PYLOM Testsuite GPOD.
#
# Last revision: 26/02/2025
from __future__ import print_function, division
import mpi4py
import pyLOM.GPOD

mpi4py.rc.recv_mprobe = False

import sys, json, numpy as np
import pyLOM

## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"').lower())

## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE, ptable=m.partition_table)
X = d.X(*VARIABLES)
t = d.get_variable('time')


##########################################
# Reconstruct Gappy Vector
##########################################
# Select snapshot and create gappy vector
snap = PARAMS['snap']
np.random.seed(5)
snapshot_POD = np.delete(X, snap, axis=1)  # Remove snapshot from training data
velox_gappy = pyLOM.GPOD.utils.set_random_elements_to_zero(X[:, snap], 99.9)

# Perform Gappy POD
gappy_model = pyLOM.GPOD.GappyPOD(**PARAMS['gpod'])
gappy_model.fit(snapshot_POD)  # Fit model
velox_recons = gappy_model.predict(velox_gappy)  # Reconstruct gappy vector

# Compute and display metrics
mae_vector  = pyLOM.math.MAE(np.ascontiguousarray(X[:,snap]), velox_recons)
rmse_vector = pyLOM.math.RMSE(np.ascontiguousarray(X[:,snap]), velox_recons, relative=False)
r2_vector   = pyLOM.math.r2(np.ascontiguousarray(X[:,snap]), velox_recons)


## Testsuite output
pyLOM.pprint(0,'TSUITE VELOX       =',velox_recons.min(),velox_recons.max(),velox_recons.mean())
pyLOM.pprint(0,'TSUITE MAE  VECTOR = %e'%mae_vector)
pyLOM.pprint(0,'TSUITE RMSE VECTOR = %e'%rmse_vector)
pyLOM.pprint(0,'TSUITE R2   VECTOR = %e'%r2_vector)


## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Python interface for POD.
#
# Last rev: 09/07/2021
from __future__ import print_function, division

cimport cython
cimport numpy as np

import numpy as np

from libc.stdlib     cimport malloc, free
from libc.string     cimport memcpy
from libc.time       cimport time
from ..vmmath.cfuncs cimport real
from ..vmmath.cfuncs cimport c_svector_norm, c_smatmul, c_svecmat, c_stemporal_mean, c_ssubtract_mean, c_stsqr_svd, c_srandomized_svd, c_scompute_truncation_residual, c_scompute_truncation
from ..vmmath.cfuncs cimport c_dvector_norm, c_dmatmul, c_dvecmat, c_dtemporal_mean, c_dsubtract_mean, c_dtsqr_svd, c_drandomized_svd, c_dcompute_truncation_residual, c_dcompute_truncation

from ..utils.cr       import cr, cr_start, cr_stop
from ..utils.errors   import raiseError


## POD run method
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _srun(float[:,:] X, int remove_mean, int randomized, int r, int q, int seed):
	'''
	Run POD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow

	Returns:
		- U:  are the POD modes.
		- S:  are the singular values.
		- V:  are the right singular vectors.
	'''
	# Variables
	cdef int m = X.shape[0], n = X.shape[1], mn = min(m,n), retval
	cdef float *X_mean
	cdef float *Y
	# Output arrays
	r = r if randomized else mn
	cdef np.ndarray[np.float32_t,ndim=2] U = np.zeros((m,r),dtype=np.float32) 
	cdef np.ndarray[np.float32_t,ndim=1] S = np.zeros((r,) ,dtype=np.float32) 
	cdef np.ndarray[np.float32_t,ndim=2] V = np.zeros((r,n),dtype=np.float32) 
	# Allocate memory
	Y = <float*>malloc(m*n*sizeof(float))
	if remove_mean:
		cr_start('POD.temporal_mean',0)
		X_mean = <float*>malloc(m*sizeof(float))
		# Compute temporal mean
		c_stemporal_mean(X_mean,&X[0,0],m,n)
		# Compute substract temporal mean
		c_ssubtract_mean(Y,&X[0,0],X_mean,m,n)
		free(X_mean)
		cr_stop('POD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],m*n*sizeof(float))
	# Compute SVD
	cr_start('POD.SVD',0)
	if randomized:
		retval = c_srandomized_svd(&U[0,0],&S[0],&V[0,0],Y,m,n,r,q,seed)
	else:
		retval = c_stsqr_svd(&U[0,0],&S[0],&V[0,0],Y,m,n)
	cr_stop('POD.SVD',0)
	free(Y)
	# Return
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _drun(double[:,:] X, int remove_mean, int randomized, int r, int q, int seed):
	'''
	Run POD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow

	Returns:
		- U:  are the POD modes.
		- S:  are the singular values.
		- V:  are the right singular vectors.
	'''
	# Variables
	cdef int m = X.shape[0], n = X.shape[1], mn = min(m,n), retval
	cdef double *X_mean
	cdef double *Y
	# Output arrays
	r = r if randomized else mn
	cdef np.ndarray[np.double_t,ndim=2] U = np.zeros((m,r),dtype=np.double) 
	cdef np.ndarray[np.double_t,ndim=1] S = np.zeros((r,) ,dtype=np.double) 
	cdef np.ndarray[np.double_t,ndim=2] V = np.zeros((r,n),dtype=np.double) 
	# Allocate memory
	Y = <double*>malloc(m*n*sizeof(double))
	if remove_mean:
		cr_start('POD.temporal_mean',0)
		X_mean = <double*>malloc(m*sizeof(double))
		# Compute temporal mean
		c_dtemporal_mean(X_mean,&X[0,0],m,n)
		# Compute substract temporal mean
		c_dsubtract_mean(Y,&X[0,0],X_mean,m,n)
		free(X_mean)
		cr_stop('POD.temporal_mean',0)
	else:
		memcpy(Y,&X[0,0],m*n*sizeof(double))
	# Compute SVD
	cr_start('POD.SVD',0)
	if randomized:
		retval = c_drandomized_svd(&U[0,0],&S[0],&V[0,0],Y,m,n,r,q,seed)
	else:
		retval = c_dtsqr_svd(&U[0,0],&S[0],&V[0,0],Y,m,n)
	cr_stop('POD.SVD',0)
	free(Y)
	# Return
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cr('POD.run')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def run(real[:,:] X, int remove_mean=True, int randomized=False, const int r=1, const int q=3, const int seed=-1):
	'''
	Run POD analysis of a matrix X.

	Inputs:
		- X[ndims*nmesh,n_temp_snapshots]: data matrix
		- remove_mean:                     whether or not to remove the mean flow

	Returns:
		- U:  are the POD modes.
		- S:  are the singular values.
		- V:  are the right singular vectors.
	'''
	seed = <int>time(NULL) if seed < 0 else seed
	if real is double:
		return _drun(X,remove_mean, randomized, r, q, seed)
	else:
		return _srun(X,remove_mean, randomized, r, q, seed)


## POD truncate method
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _struncate(float[:,:] U, float[:] S, float[:,:] V, float r):
	'''
	Truncate POD matrices (U,S,V) given a residual r.

	Inputs:
		- U(m,n)  are the POD modes.
		- S(n)    are the singular values.
		- V(n,n)  are the right singular vectors.
		- r       target residual, number of modes, or cumulative energy threshold.
					* If r >= 1, it is treated as the number of modes.
					* If r < 1 and r > 0 it is treated as the residual target.
					* If r < 1 and r < 0 it is treated as the fraction of cumulative energy to retain.
					Note:  must be in (0,-1] and r = -1 is valid

	Returns:
		- U(m,N)  are the POD modes (truncated at N).
		- S(N)    are the singular values (truncated at N).
		- V(N,n)  are the right singular vectors (truncated at N).
	'''
	cdef int m = U.shape[0], n = V.shape[1], nmod = U.shape[1], N
	# Compute N using S
	N  = int(r) if r >=1 else c_scompute_truncation_residual(&S[0],r,n)
	# Allocate output arrays
	cdef np.ndarray[np.float32_t,ndim=2] Ur = np.zeros((m,N),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] Sr = np.zeros((N,), dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] Vr = np.zeros((N,n),dtype=np.float32)
	# Truncate
	c_scompute_truncation(&Ur[0,0],&Sr[0],&Vr[0,0],&U[0,0],&S[0],&V[0,0],m,n,nmod,N)
	# Return
	return Ur, Sr, Vr

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dtruncate(double[:,:] U, double[:] S, double[:,:] V, double r):
	'''
	Truncate POD matrices (U,S,V) given a residual r.

	Inputs:
		- U(m,n)  are the POD modes.
		- S(n)    are the singular values.
		- V(n,n)  are the right singular vectors.
		- r       target residual, number of modes, or cumulative energy threshold.
					* If r >= 1, it is treated as the number of modes.
					* If r < 1 and r > 0 it is treated as the residual target.
					* If r < 1 and r < 0 it is treated as the fraction of cumulative energy to retain.
					Note:  must be in (0,-1] and r = -1 is valid

	Returns:
		- U(m,N)  are the POD modes (truncated at N).
		- S(N)    are the singular values (truncated at N).
		- V(N,n)  are the right singular vectors (truncated at N).
	'''
	cdef int m = U.shape[0], n = V.shape[1], N, nmod = U.shape[1]
	# Compute N using S
	N  = int(r) if r >=1 else c_dcompute_truncation_residual(&S[0],r,n)
	# Allocate output arrays
	cdef np.ndarray[np.double_t,ndim=2] Ur = np.zeros((m,N),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] Sr = np.zeros((N,), dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] Vr = np.zeros((N,n),dtype=np.double)
	# Truncate
	c_dcompute_truncation(&Ur[0,0],&Sr[0],&Vr[0,0],&U[0,0],&S[0],&V[0,0],m,n,nmod,N)
	# Return
	return Ur, Sr, Vr

@cr('POD.truncate')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def truncate(real[:,:] U, real[:] S, real[:,:] V, real r=1e-8):
	'''
	Truncate POD matrices (U,S,V) given a residual r.

	Inputs:
		- U(m,n)  are the POD modes.
		- S(n)    are the singular values.
		- V(n,n)  are the right singular vectors.
		- r       target residual, number of modes, or cumulative energy threshold.
					* If r >= 1, it is treated as the number of modes.
					* If r < 1 and r > 0 it is treated as the residual target.
					* If r < 1 and r < 0 it is treated as the fraction of cumulative energy to retain.
					Note:  must be in (0,-1] and r = -1 is valid

	Returns:
		- U(m,N)  are the POD modes (truncated at N).
		- S(N)    are the singular values (truncated at N).
		- V(N,n)  are the right singular vectors (truncated at N).
	'''
	if real is double:
		return _dtruncate(U,S,V,r)
	else:
		return _struncate(U,S,V,r)


## POD reconstruct method
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sreconstruct(float[:,:] U, float[:] S, float[:,:] V):
	'''
	Reconstruct the flow given the POD decomposition matrices
	that can be possibly truncated.
	N is the truncated size
	n is the number of snapshots

	Inputs:
		- U(m,N)  are the POD modes.
		- S(N)    are the singular values.
		- V(N,n)  are the right singular vectors.

	Outputs
		- X(m,n)  is the reconstructed flow.
	'''
	cdef int m = U.shape[0], N = S.shape[0], n = V.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] X = np.zeros((m,n),dtype=np.float32)
	cdef float *Vtmp
	# Copy V to Vtmp so V is not modified by the routine
	Vtmp = <float*>malloc(N*n*sizeof(float))
	memcpy(Vtmp,&V[0,0],N*n*sizeof(float))
	# Scale V by S doing V' = diag(S) x V
	c_svecmat(&S[0],Vtmp,N,n)
	# Compute X = U x V'
	c_smatmul(&X[0,0],&U[0,0],Vtmp,m,n,N)
	# Return
	free(Vtmp)
	return X

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dreconstruct(double[:,:] U, double[:] S, double[:,:] V):
	'''
	Reconstruct the flow given the POD decomposition matrices
	that can be possibly truncated.
	N is the truncated size
	n is the number of snapshots

	Inputs:
		- U(m,N)  are the POD modes.
		- S(N)    are the singular values.
		- V(N,n)  are the right singular vectors.

	Outputs
		- X(m,n)  is the reconstructed flow.
	'''
	cdef int m = U.shape[0], N = S.shape[0], n = V.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] X = np.zeros((m,n),dtype=np.double)
	cdef double *Vtmp
	# Copy V to Vtmp so V is not modified by the routine
	Vtmp = <double*>malloc(N*n*sizeof(double))
	memcpy(Vtmp,&V[0,0],N*n*sizeof(double))
	# Scale V by S doing V' = diag(S) x V
	c_dvecmat(&S[0],Vtmp,N,n)
	# Compute X = U x V'
	c_dmatmul(&X[0,0],&U[0,0],Vtmp,m,n,N)
	# Return
	free(Vtmp)
	return X

@cr('POD.reconstruct')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def reconstruct(real[:,:] U, real[:] S, real[:,:] V):
	'''
	Reconstruct the flow given the POD decomposition matrices
	that can be possibly truncated.
	N is the truncated size
	n is the number of snapshots

	Inputs:
		- U(m,N)  are the POD modes.
		- S(N)    are the singular values.
		- V(N,n)  are the right singular vectors.

	Outputs
		- X(m,n)  is the reconstructed flow.
	'''
	if real is double:
		return _dreconstruct(U,S,V)
	else:
		return _sreconstruct(U,S,V)



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# POD general utilities.
#
# Last rev: 29/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..          import inp_out as io
from ..utils     import cr_nvtx as cr, gpu_to_cpu


@cr('POD.extract_modes')
def extract_modes(U,ivar,npoints,modes=[],reshape=True):
	'''
	Extract modes for a certain variables
	'''
	p = cp if type(U) is cp.ndarray else np
	nvars = U.shape[0]//npoints
	# Define modes to extract
	if len(modes) == 0: modes = p.arange(1,U.shape[1]+1,dtype=p.int32)
	# Allocate output array
	out =p.zeros((npoints,len(modes)),U.dtype)
	for i,m in enumerate(modes):
		out[:,i] = U[ivar-1:nvars*npoints:nvars,m-1]
	# Return reshaped output
	return out.reshape((len(modes)*npoints,),order='C') if reshape else out


@cr('POD.save')
def save(fname,U,S,V,ptable,nvars=1,pointData=True,mode='w'):
	'''
	Store POD variables in serial or parallel
	according to the partition used to compute the POD.
	'''
	io.h5_save_POD(fname,gpu_to_cpu(U),gpu_to_cpu(S),gpu_to_cpu(V),ptable,nvars=nvars,pointData=pointData,mode=mode)


@cr('POD.load')
def load(fname,vars=['U','S','V'],nmod=-1,ptable=None):
	'''
	Load POD variables in serial or parallel
	according to the partition used to compute the POD.
	'''
	return io.h5_load_POD(fname,vars,nmod,ptable)



# Configuration file for the Sphinx documentation builder.
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))
# import pyLOM
# -- Project information -----------------------------------------------------
project = 'pyLOM'
copyright = '2023-2025'
author = 'pyLOM developers'
# TODO: ask to add pyLOM.__version__  to avoid this hardcoded variable
version = '2.1.0' # pyLOM.__version__ is not defined

# -- General configuration ---------------------------------------------------
extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',
    'sphinx.ext.mathjax',
    'sphinx_design',
    'nbsphinx'
]

# Set the default syntax highlighter
highlight_language = 'python'

# For nbsphinx
nbsphinx_codecell_lexer = 'ipython3'

templates_path = ['_templates']
exclude_patterns = []

# -- Options for autodoc ----------------------------------------------------
autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'show-inheritance': True,
    # 'special-members': '__init__',
    'imported-members': True, # To show the modules imported on the __init__.py of every package
    'ignore-module-all': True,  # If True, ignore __all__ variable defined in the package's __init__.py
    'undoc-members': False, # If True, include members without docstrings
    'exclude-members': '__weakref__'
}
# without this, the undocumented members are still shown, even with undoc-members=False
autodoc_member_order = 'bysource'
autodoc_preserve_defaults = True


# -- Options for Napoleon --------------------------------------------------
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False # don't document privete methods (the ones that start with _)
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_type_aliases = None

# -- Intersphinx configuration --------------------------------------------
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

# -- Options for HTML output ----------------------------------------------
html_theme = 'pydata_sphinx_theme'
html_theme_options = {
    "github_url": "https://github.com/ArnauMiro/pyLowOrder",
    "use_edit_page_button": True,
    "show_nav_level": 2,
    "show_toc_level": 2,
    "navigation_depth": 4,
    "collapse_navigation": False,
    "logo": {
        "alt_text": "pyLOM",
        "image_light": "_static/logo_tmp.webp",
        "image_dark": "_static/logo_tmp.webp",
    },
}

# These folders are copied to the documentation's HTML output
html_static_path = ['_static']
html_css_files = [
    'custom.css',
]
html_favicon = '_static/favicon_tmp.ico'
# -- Options for HTML theme -----------------------------------------------
html_context = {
    "github_user": "ArnauMiro",
    "github_repo": "pyLowOrder",
    "github_version": "master",
    "doc_path": "docs/source",
}


#!/usr/bin/env python
#
# pyLOM, utils.
#
# NVTX profiler
#
# Last rev: 12/03/2025
from __future__ import print_function, division

import functools

try:
	import nvtx

	def nvtxp(ch_name,color="blue"):
		'''
		CR NVTX decorator
		'''
		def decorator(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				with nvtx.annotate(message=ch_name,color=color):
					out = func(*args,**kwargs)
				return out
			return wrapper
		return decorator

except:
	def nvtxp(ch_name,color="blue"):
		'''
		CR NVTX decorator
		'''
		def decorator(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				out = func(*args,**kwargs)
				return out
			return wrapper
		return decorator


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# SPOD Module
#
# Last rev: 06/10/2022

from .wrapper import run
from .utils   import extract_modes, save, load
from .plots import plotMode, plotSpectra
